from typing import Callable, List, Literal, Optional, Tuple, Union

import numpy as np
from PIL import Image

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.metrics.image_data_matching import ImageDataMatching
from cv_pipeliner.utils.images import concat_images
from cv_pipeliner.visualizers.core.image_data import visualize_image_data


def get_true_and_pred_images_data_with_visualized_labels(
    image_data_matching: ImageDataMatching, error_type: Literal["detection", "pipeline"], label: str = None
) -> ImageData:
    """
    Create true and pred ImageData with changed label for visualization

    For detection error_type, the label will be one of ["TP", "FP", "FN"]

    For pipeline error_type, the label will be in format "label {matching_error_type}"
    (where matching_error_type is one of ["TP", "FP", "FN", "TP (extra bbox)", "FP (extra bbox)"])
    """
    for tag, tag_image_data in [
        ("true", image_data_matching.true_image_data),
        ("pred", image_data_matching.pred_image_data),
    ]:
        tag_bboxes_data_with_visualized_label = []
        for tag_bbox_data_matching in image_data_matching.bboxes_data_matchings:
            if tag == "true":
                tag_bbox_data = tag_bbox_data_matching.true_bbox_data
            elif tag == "pred":
                tag_bbox_data = tag_bbox_data_matching.pred_bbox_data

            if tag_bbox_data is None:
                continue

            if error_type == "detection":
                label_caption = f"[{tag_bbox_data_matching.get_detection_error_type(label=label)}]"
            elif error_type == "pipeline":
                pipeline_error_type = tag_bbox_data_matching.get_pipeline_error_type(label=label)
                if label is None:
                    label_caption = f"{tag_bbox_data.label} [{pipeline_error_type}]"
                else:
                    label_caption = f"{tag_bbox_data.label} [{pipeline_error_type}, cls. '{label}']"

            tag_bbox_data_with_visualized_label = BboxData(
                image_path=tag_bbox_data.image_path,
                cropped_image=tag_bbox_data.cropped_image,
                xmin=tag_bbox_data.xmin,
                ymin=tag_bbox_data.ymin,
                xmax=tag_bbox_data.xmax,
                ymax=tag_bbox_data.ymax,
                detection_score=tag_bbox_data.detection_score,
                label=label_caption,
                classification_score=tag_bbox_data.classification_score,
            )
            tag_bboxes_data_with_visualized_label.append(tag_bbox_data_with_visualized_label)

        tag_image_data_with_visualized_labels = ImageData(
            image_path=tag_image_data.image_path,
            image=tag_image_data.image,
            bboxes_data=tag_bboxes_data_with_visualized_label,
        )
        if tag == "true":
            true_image_data_with_visualized_labels = tag_image_data_with_visualized_labels
        elif tag == "pred":
            pred_image_data_with_visualized_labels = tag_image_data_with_visualized_labels

    return true_image_data_with_visualized_labels, pred_image_data_with_visualized_labels


error_type = Literal["TP", "FP", "FN", "TP (extra bbox)", "FP (extra bbox)"]


def visualize_image_data_matching_side_by_side(
    image_data_matching: ImageDataMatching,
    error_type: Literal["detection", "pipeline"],
    true_include_labels: bool = False,
    pred_include_labels: bool = False,
    true_include_mask: bool = False,
    pred_include_mask: bool = False,
    true_score_type: Literal["detection", "classification"] = None,
    pred_score_type: Literal["detection", "classification"] = None,
    true_filter_by_error_types: List[error_type] = ["TP", "FP", "FN", "TP (extra bbox)", "FP (extra bbox)"],
    pred_filter_by_error_types: List[error_type] = ["TP", "FP", "FN", "TP (extra bbox)", "FP (extra bbox)"],
    known_labels: Optional[List[str]] = None,
    draw_base_labels_with_given_label_to_base_label_image: Callable[[str], np.ndarray] = None,
    keypoints_radius: int = 5,
    include_additional_bboxes_data: bool = False,
    additional_bboxes_data_depth: Optional[int] = None,
    fontsize: int = 24,
    thickness: int = 4,
    mask_alpha: float = 0.5,
    thumbnail_size: Optional[Union[int, Tuple[int, int]]] = None,
    label: str = None,
    return_as_pil_image: bool = False,
    exif_transpose: bool = False,
) -> Union[np.ndarray, Image.Image]:
    (
        true_image_data_with_visualized_labels,
        pred_image_data_with_visualized_labels,
    ) = get_true_and_pred_images_data_with_visualized_labels(
        image_data_matching=image_data_matching, error_type=error_type, label=label
    )

    true_visualized_labels = [bbox_data.label for bbox_data in true_image_data_with_visualized_labels.bboxes_data]
    pred_visualized_labels = [bbox_data.label for bbox_data in pred_image_data_with_visualized_labels.bboxes_data]
    true_filter_by_labels = [
        label
        for label in true_visualized_labels
        if any(f"[{matching_error_type}]" in label for matching_error_type in true_filter_by_error_types)
    ]
    pred_filter_by_labels = [
        label
        for label in pred_visualized_labels
        if any(f"[{matching_error_type}]" in label for matching_error_type in pred_filter_by_error_types)
    ]
    true_image = visualize_image_data(
        image_data=true_image_data_with_visualized_labels,
        include_labels=true_include_labels,
        score_type=true_score_type,
        filter_by_labels=true_filter_by_labels,
        known_labels=known_labels,
        include_mask=true_include_mask,
        mask_alpha=mask_alpha,
        draw_base_labels_with_given_label_to_base_label_image=draw_base_labels_with_given_label_to_base_label_image,
        keypoints_radius=keypoints_radius,
        include_additional_bboxes_data=include_additional_bboxes_data,
        additional_bboxes_data_depth=additional_bboxes_data_depth,
        fontsize=fontsize,
        thickness=thickness,
        thumbnail_size=thumbnail_size,
        return_as_pil_image=False,
        exif_transpose=exif_transpose,
    )
    pred_image = visualize_image_data(
        image_data=pred_image_data_with_visualized_labels,
        include_labels=pred_include_labels,
        score_type=pred_score_type,
        filter_by_labels=pred_filter_by_labels,
        known_labels=known_labels,
        include_mask=pred_include_mask,
        mask_alpha=mask_alpha,
        draw_base_labels_with_given_label_to_base_label_image=draw_base_labels_with_given_label_to_base_label_image,
        keypoints_radius=keypoints_radius,
        include_additional_bboxes_data=include_additional_bboxes_data,
        additional_bboxes_data_depth=additional_bboxes_data_depth,
        fontsize=fontsize,
        thickness=thickness,
        thumbnail_size=thumbnail_size,
        return_as_pil_image=False,
    )
    image = concat_images(true_image, pred_image, how="horizontally")
    if return_as_pil_image:
        return Image.fromarray(image)
    return image



import json
import tempfile
from pathlib import Path
from typing import Callable, List, Literal, Tuple, Type, Union

import cv2
import fsspec
import numpy as np
from pathy import Pathy

from cv_pipeliner.core.inference_model import get_preprocess_input_from_script_file
from cv_pipeliner.inference_models.detection.core import (
    DetectionInput,
    DetectionModel,
    DetectionModelSpec,
    DetectionOutput,
)


class Detectron2_ModelSpec(DetectionModelSpec):
    input_size: Union[Tuple[int, int], List[int]]
    model_path: Union[str, Pathy]  # can be also tf.keras.Model

    bboxes_output_index: int
    scores_output_index: int
    classes_output_index: int
    input_format: Literal["RGB", "BGR"]
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    keypoints_output_index: Union[int, None] = None
    keypoints_heatmap_index: Union[int, None] = None
    class_names: Union[List[str], str, Path, None] = None
    device: Literal["cpu", "cuda"] = "cpu"
    input_type: Literal["detectron2", "caffe2"] = "detectron2"

    @property
    def inference_model_cls(self) -> Type["Detectron2_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.detectron2 import (
            Detectron2_DetectionModel,
        )

        return Detectron2_DetectionModel


def heatmaps_to_keypoints(maps: "torch.Tensor", rois: "torch.Tensor") -> "torch.Tensor":  # noqa: F821
    """
    Extract predicted keypoint locations from heatmaps.
    Args:
        maps (Tensor): (#ROIs, #keypoints, POOL_H, POOL_W). The predicted heatmap of logits for
            each ROI and each keypoint.
        rois (Tensor): (#ROIs, 4). The box of each ROI.
    Returns:
        Tensor of shape (#ROIs, #keypoints, 4) with the last dimension corresponding to
        (x, y, logit, score) for each keypoint.
    When converting discrete pixel indices in an NxN image to a continuous keypoint coordinate,
    we maintain consistency with :meth:`Keypoints.to_heatmap` by using the conversion from
    Heckbert 1990: c = d + 0.5, where d is a discrete coordinate and c is a continuous coordinate.
    """
    import torch
    from torch.nn import functional as F

    # The decorator use of torch.no_grad() was not supported by torchscript.
    # https://github.com/pytorch/pytorch/issues/44768
    maps = maps.detach()
    rois = rois.detach()

    offset_x = rois[:, 0]
    offset_y = rois[:, 1]

    widths = (rois[:, 2] - rois[:, 0]).clamp(min=1)
    heights = (rois[:, 3] - rois[:, 1]).clamp(min=1)
    widths_ceil = widths.ceil()
    heights_ceil = heights.ceil()

    num_rois, num_keypoints = maps.shape[:2]
    xy_preds = maps.new_zeros(rois.shape[0], num_keypoints, 4)

    width_corrections = widths / widths_ceil
    height_corrections = heights / heights_ceil

    keypoints_idx = torch.arange(num_keypoints, device=maps.device)

    for i in range(num_rois):
        outsize = (int(heights_ceil[i]), int(widths_ceil[i]))
        roi_map = F.interpolate(maps[[i]], size=outsize, mode="bicubic", align_corners=False).squeeze(
            0
        )  # #keypoints x H x W

        # softmax over the spatial region
        max_score, _ = roi_map.view(num_keypoints, -1).max(1)
        max_score = max_score.view(num_keypoints, 1, 1)
        tmp_full_resolution = (roi_map - max_score).exp_()
        tmp_pool_resolution = (maps[i] - max_score).exp_()
        # Produce scores over the region H x W, but normalize with POOL_H x POOL_W,
        # so that the scores of objects of different absolute sizes will be more comparable
        roi_map_scores = tmp_full_resolution / tmp_pool_resolution.sum((1, 2), keepdim=True)

        w = roi_map.shape[2]
        pos = roi_map.view(num_keypoints, -1).argmax(1)

        x_int = pos % w
        y_int = (pos - x_int) // w

        assert (roi_map_scores[keypoints_idx, y_int, x_int] == roi_map_scores.view(num_keypoints, -1).max(1)[0]).all()

        x = (x_int.float() + 0.5) * width_corrections[i]
        y = (y_int.float() + 0.5) * height_corrections[i]

        xy_preds[i, :, 0] = x + offset_x[i]
        xy_preds[i, :, 1] = y + offset_y[i]
        xy_preds[i, :, 2] = roi_map[keypoints_idx, y_int, x_int]
        xy_preds[i, :, 3] = roi_map_scores[keypoints_idx, y_int, x_int]

    return xy_preds


class Detectron2_DetectionModel(DetectionModel):
    def _load_pt_model(self, model_spec: Detectron2_ModelSpec):
        import torch

        temp_dir = tempfile.TemporaryDirectory()
        temp_dir_path = Path(temp_dir.name)
        model_config_path = temp_dir_path / Pathy.fluid(model_spec.model_path).name
        with open(model_config_path, "wb") as out:
            with fsspec.open(model_spec.model_path, "rb") as src:
                out.write(src.read())
        self.model = torch.jit.load(model_config_path).to(model_spec.device)
        self.model.eval()
        temp_dir.cleanup()

    def __init__(self, model_spec: Detectron2_ModelSpec):
        super().__init__(model_spec)

        if model_spec.class_names is not None:
            if isinstance(model_spec.class_names, str) or isinstance(model_spec.class_names, Path):
                with fsspec.open(model_spec.class_names, "r", encoding="utf-8") as out:
                    self.class_names = np.array(json.load(out))
            else:
                self.class_names = np.array(model_spec.class_names)
        else:
            self.class_names = None

        if isinstance(model_spec, Detectron2_ModelSpec):
            self._load_pt_model(model_spec)
            self.device = model_spec.device
            self.input_format = model_spec.device
            self.input_type = model_spec.input_type
        else:
            raise ValueError(f"{Detectron2_DetectionModel.__name__} got unknown DetectionModelSpec: {type(model_spec)}")

        if isinstance(model_spec.preprocess_input, str) or isinstance(model_spec.preprocess_input, Path):
            self._preprocess_input = get_preprocess_input_from_script_file(script_file=model_spec.preprocess_input)
        else:
            if model_spec.preprocess_input is None:
                self._preprocess_input = lambda x: x
            else:
                self._preprocess_input = model_spec.preprocess_input

    def _raw_predict_single_image(self, image: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        import torch

        if self.input_format == "BGR":
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        image = torch.tensor(image).permute(2, 1, 0).to(self.device)

        if self.input_type == "detectron2":
            predictions = self.model(image)
        elif self.input_type == "caffe2":
            image = image[None, ...]
            im_info = torch.tensor([[*self.input_size, 1.0]])
            predictions = self.model((image, im_info))

        raw_bboxes = predictions[self.model_spec.bboxes_output_index]
        if self.model_spec.keypoints_output_index is None:
            if self.model_spec.keypoints_heatmap_index is not None:
                raw_keypoints_heatmaps = predictions[self.model_spec.keypoints_heatmap_index]
                raw_keypoints = heatmaps_to_keypoints(raw_keypoints_heatmaps, raw_bboxes)
            else:
                raw_keypoints = np.array([]).reshape(len(raw_bboxes), 0, 2)
        else:
            raw_keypoints = predictions[self.model_spec.keypoints_output_index]

        raw_bboxes = raw_bboxes.detach().cpu().numpy()
        raw_keypoints = raw_keypoints.detach().cpu().numpy()[:, :, :2]
        raw_scores = predictions[self.model_spec.scores_output_index].detach().cpu().numpy()
        raw_classes = predictions[self.model_spec.classes_output_index].detach().cpu().numpy()

        return raw_bboxes, raw_keypoints, raw_scores, raw_classes

    def _postprocess_prediction(
        self,
        raw_bboxes: np.ndarray,
        raw_keypoints: np.ndarray,
        raw_scores: np.ndarray,
        raw_classes: np.ndarray,
        score_threshold: float,
        classification_top_n: int,
        height: int,
        width: int,
    ) -> Tuple[
        List[Tuple[int, int, int, int]], List[List[Tuple[int, int]]], List[float], List[List[str]], List[List[float]]
    ]:
        raw_bboxes[:, [0, 2]] = raw_bboxes[:, [0, 2]] / self.input_size[0] * width
        raw_bboxes[:, [1, 3]] = raw_bboxes[:, [1, 3]] / self.input_size[1] * height
        raw_bboxes = raw_bboxes.round().astype(int)
        raw_keypoints[:, :, 0] = raw_keypoints[:, :, 0] / self.input_size[0] * width
        raw_keypoints[:, :, 1] = raw_keypoints[:, :, 1] / self.input_size[1] * height
        raw_keypoints = raw_keypoints.round().astype(int)
        mask = raw_scores > score_threshold
        bboxes = raw_bboxes[mask]
        keypoints = raw_keypoints[mask]
        scores = raw_scores[mask]
        classes = raw_classes[mask]

        correct_non_repeated_bboxes_idxs = []
        bboxes_set = set()
        for idx, bbox in enumerate(bboxes):
            xmin, ymin, xmax, ymax = bbox
            if xmax - xmin > 0 and ymax - ymin > 0 and (xmin, ymin, xmax, ymax) not in bboxes_set:
                bboxes_set.add((xmin, ymin, xmax, ymax))
                correct_non_repeated_bboxes_idxs.append(idx)

        bboxes = bboxes[correct_non_repeated_bboxes_idxs]
        keypoints = keypoints[correct_non_repeated_bboxes_idxs]
        scores = scores[correct_non_repeated_bboxes_idxs]
        classes = classes[correct_non_repeated_bboxes_idxs]
        classes_scores = scores.copy()
        if self.class_names is not None:
            class_names_top_n = np.array(
                [
                    [class_name for i in range(classification_top_n)]
                    for class_name in self.class_names[(classes.astype(np.int32))]
                ]
            )
            classes_scores_top_n = np.array([[score for _ in range(classification_top_n)] for score in classes_scores])
        else:
            class_names_top_n = np.array([[None for _ in range(classification_top_n)] for _ in classes])
            classes_scores_top_n = np.array([[score for _ in range(classification_top_n)] for score in classes_scores])

        return bboxes, keypoints, scores, class_names_top_n, classes_scores_top_n

    def predict(self, input: DetectionInput, score_threshold: float, classification_top_n: int = 1) -> DetectionOutput:
        (n_pred_bboxes, n_pred_keypoints, n_pred_scores, n_pred_class_names_top_k, n_pred_scores_top_k) = (
            [],
            [],
            [],
            [],
            [],
        )

        for image in input:
            height, width, _ = image.shape
            resized_image = self.preprocess_input([image])[0]
            raw_bboxes, raw_keypoints, raw_scores, raw_classes = self._raw_predict_single_image(resized_image)
            bboxes, keypoints, scores, class_names_top_k, classes_scores_top_k = self._postprocess_prediction(
                raw_bboxes=raw_bboxes,
                raw_keypoints=raw_keypoints,
                raw_scores=raw_scores,
                raw_classes=raw_classes,
                score_threshold=score_threshold,
                classification_top_n=classification_top_n,
                width=width,
                height=height,
            )

            n_pred_bboxes.append(bboxes)
            n_pred_keypoints.append(keypoints)
            n_pred_scores.append(scores)
            n_pred_class_names_top_k.append(class_names_top_k)
            n_pred_scores_top_k.append(classes_scores_top_k)

        n_pred_masks = [[[] for _ in pred_bboxes] for pred_bboxes in n_pred_bboxes]
        return (
            n_pred_bboxes,
            n_pred_keypoints,
            n_pred_masks,
            n_pred_scores,
            n_pred_class_names_top_k,
            n_pred_scores_top_k,
        )

    def preprocess_input(self, input: DetectionInput):
        return self._preprocess_input(input)

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.model_spec.input_size



import json
from pathlib import Path
from typing import IO, Any, Callable, Dict, Iterator, List, Optional, Tuple, Type, Union

import fsspec
import numpy as np
import pandas as pd

# from threading import Semaphore
from datapipe.run_config import RunConfig
from datapipe.store.database import DBConn, TableStoreDB
from datapipe.store.filedir import ItemStoreFileAdapter, TableStoreFiledir
from datapipe.store.table_store import TableStore
from datapipe.types import (
    DataDF,
    DataSchema,
    IndexDF,
    MetaSchema,
    data_to_index,
    index_difference,
    index_intersection,
    index_to_data,
)
from sqlalchemy import JSON, Column, Integer, String

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.utils.fiftyone import FifyOneSession
from cv_pipeliner.utils.images_datas import get_all_bboxes_data_in_image_data
from cv_pipeliner.utils.imagesize import get_image_size


class ImageDataFile(ItemStoreFileAdapter):
    """
    Converts each ImageData file into Pandas record
    """

    mode = "t"

    def __init__(
        self,
        image_data_cls: Type[ImageData] = ImageData,
    ):
        self.image_data_cls = image_data_cls

    def load(self, f: IO) -> ImageData:
        image_data_json = json.load(f)
        image_data = self.image_data_cls.from_json(image_data_json) if image_data_json is not None else None
        return {"image_data": image_data}

    def dump(self, obj: Dict[str, Any], f: IO) -> None:
        image_data: ImageData = obj["image_data"]
        f.write(image_data.json(indent=4, ensure_ascii=False))


class NumpyDataFile(ItemStoreFileAdapter):
    """
    Converts each npy file into Pandas record
    """

    mode = "b"

    def load(self, f: IO) -> np.ndarray:
        ndarray = np.load(f)
        return {"ndarray": ndarray}

    def dump(self, obj: Dict[str, Any], f: IO) -> None:
        ndarray: np.ndarray = obj["ndarray"]
        return np.save(f, ndarray)


class GetImageSizeFile(ItemStoreFileAdapter):
    mode = "b"

    def load(self, f: IO) -> Dict[str, Tuple[int, int]]:
        image_size = get_image_size(f)
        return {"image_size": image_size}

    def dump(self, obj: Dict[str, Tuple[int, int]], f: IO) -> None:
        raise NotImplementedError


class YOLOLabelsFile(ItemStoreFileAdapter):
    mode = "b"

    def __init__(self, class_names: List[str], img_format: str):
        from cv_pipeliner.data_converters.yolo import YOLODataConverter

        self.yolo_converter = YOLODataConverter(class_names)
        self.img_format = img_format

    def load(self, f: fsspec.core.OpenFile) -> Dict[str, Tuple[int, int]]:
        filepath = Path(f.path)
        assert filepath.parent.name == "labels"
        image_path = filepath.parent.parent / "images" / f"{filepath.stem}.{self.img_format}"
        if f.fs.protocol in ["file"] or f.fs.protocol is None:
            prefix = ""
        else:
            protocol = f.fs.protocol
            if isinstance(protocol, tuple):
                protocol = protocol[0]
            prefix = f"{protocol}://"
        image_data = self.yolo_converter.get_image_data_from_annot(image_path=f"{prefix}{image_path}", annot=f)
        return {"image_size": image_data}

    def dump(self, obj: Dict[str, Tuple[int, int]], f: fsspec.core.OpenFile) -> None:
        image_data: ImageData = obj["image_data"]
        yolo_data = self.yolo_converter.get_annot_from_image_data(image_data)
        f.write("\n".join(yolo_data).encode())


class ImageDataTableStoreDB(TableStoreDB):
    def __init__(
        self,
        dbconn: Union[DBConn, str],
        name: str,
        data_sql_schema: List[Column],
        create_table: bool = False,
        image_data_cls: Type[ImageData] = ImageData,
    ) -> None:
        # assert all([column.primary_key for column in data_sql_schema])
        assert "image_data" not in [column.name for column in data_sql_schema]
        self.data_sql_schema_raw = data_sql_schema
        super().__init__(
            dbconn=dbconn,
            name=name,
            data_sql_schema=data_sql_schema + [Column("image_data", JSON)],
            create_table=create_table,
        )
        self.create_table = create_table
        self.image_data_cls = image_data_cls

    def __reduce__(self) -> Tuple[Any, ...]:
        return self.__class__, (
            self.dbconn,
            self.name,
            self.data_sql_schema_raw,
            self.create_table,
            self.image_data_cls,
        )

    def insert_rows(self, df: DataDF) -> None:
        df["image_data"] = df["image_data"].apply(
            lambda image_data: json.loads(image_data.json()) if image_data is not None else None
        )
        super().insert_rows(df)

    def read_rows(self, idx: Optional[IndexDF] = None) -> pd.DataFrame:
        df = super().read_rows(idx=idx)
        df["image_data"] = df["image_data"].apply(
            lambda image_data_json: (
                self.image_data_cls.from_json(image_data_json) if image_data_json is not None else None
            )
        )
        return df


class EmptyItemStoreFileAdapter(ItemStoreFileAdapter):
    mode = "b"


class ConnectedImageDataTableStore(TableStore):
    def __init__(
        self,
        images_table_store: TableStoreFiledir,
        images_data_table_store: Union[ImageDataTableStoreDB, TableStoreFiledir],
    ) -> None:
        if isinstance(images_data_table_store, TableStoreFiledir):
            assert isinstance(images_data_table_store.adapter, ImageDataFile)
        self.images_data_store = images_data_table_store
        self.images_table_store = images_table_store
        for key in self.images_table_store.primary_keys:
            assert key in self.images_data_store.primary_keys, f"Missing key for images_data_store: {key}"
        self.attrnames = self.images_table_store.attrnames
        self.primary_schema = self.images_data_store.get_primary_schema()

    def get_primary_schema(self) -> DataSchema:
        return [column for column in self.primary_schema if column.primary_key]

    def get_meta_schema(self) -> MetaSchema:
        return []

    def delete_rows(self, idx: IndexDF) -> None:
        self.images_data_store.delete_rows(idx)

    def _set_images_paths(self, df: DataDF, force_update_meta: bool = False) -> None:
        for row_idx in df.index:
            if df.loc[row_idx, "image_data"] is None:
                continue
            idxs_values = df.loc[row_idx, self.attrnames].tolist()
            image_path_candidates = self.images_table_store._filenames_from_idxs_values(idxs_values)
            if len(image_path_candidates) > 1:
                for image_path in image_path_candidates:
                    openfile = fsspec.open(image_path)
                    if openfile.fs.exists(openfile.path):
                        break
            else:
                image_path = image_path_candidates[0]
            df.loc[row_idx, "image_data"].image_path = image_path
            for bbox_data in get_all_bboxes_data_in_image_data(df.loc[row_idx, "image_data"]):
                bbox_data.image_path = image_path
            if force_update_meta:
                df.loc[row_idx, "image_data"].get_image_size()

    def insert_rows(self, df: DataDF, force_update_meta: bool = False) -> None:
        self._set_images_paths(df, force_update_meta=force_update_meta)
        self.images_data_store.insert_rows(df)

    def update_rows(self, df: DataDF, force_update_meta: bool = False) -> None:
        if df.empty:
            return
        self.delete_rows(data_to_index(df, self.primary_keys))
        self.insert_rows(df, force_update_meta=True)

    def read_rows(self, idx: IndexDF = None) -> DataDF:
        df_images_data = self.images_data_store.read_rows(idx)
        self._set_images_paths(df_images_data)
        return df_images_data

    def read_rows_meta_pseudo_df(self, chunksize: int = 1000, run_config: RunConfig = None) -> Iterator[DataDF]:
        # FIXME сделать честную чанкированную реализацию во всех сторах
        for df_meta in self.images_data_store.read_rows_meta_pseudo_df(chunksize=chunksize, run_config=run_config):
            yield df_meta


class FiftyOneImagesDataTableStore(TableStore):
    def __init__(
        self,
        dataset: str,
        fo_session: FifyOneSession,
        fo_detections_label: Optional[str] = None,
        fo_classification_label: Optional[str] = None,
        fo_keypoints_label: Optional[str] = None,
        primary_schema: DataSchema = None,
        images_table_store: Optional[TableStoreFiledir] = None,
        mapping_filepath: Callable[[str], str] = lambda filepath: filepath,
        inverse_mapping_filepath: Callable[[str], str] = lambda filepath: filepath,
        rm_only_fo_fields: bool = True,
        additional_info_keys_in_fo_detections: List[str] = [],
        additional_info_keys_in_sample: List[str] = [],
        create_dataset_if_empty: bool = True,
        image_data_cls: Type[ImageData] = ImageData,
        bbox_data_cls: Type[BboxData] = BboxData,
    ):
        self.dataset = dataset
        self.fo_detections_label = fo_detections_label
        self.fo_classification_label = fo_classification_label
        self.fo_keypoints_label = fo_keypoints_label
        self.fo_session = fo_session
        self.images_table_store = images_table_store
        self.mapping_filepath = mapping_filepath
        self.inverse_mapping_filepath = inverse_mapping_filepath
        self.rm_only_fo_fields = rm_only_fo_fields
        self.additional_info_keys_in_fo_detections = additional_info_keys_in_fo_detections
        self.additional_info_keys_in_sample = additional_info_keys_in_sample
        self.image_data_cls = image_data_cls
        self.bbox_data_cls = bbox_data_cls
        if primary_schema is not None:
            assert all([isinstance(column.type, (String, Integer)) for column in primary_schema])
            self.primary_schema = primary_schema
        else:
            self.primary_schema = [Column("filepath", String(100), primary_key=True)]
        self.attrnames = [column.name for column in self.primary_schema]
        if self.images_table_store is not None:
            for key in self.images_table_store.primary_keys:
                assert key in self.attrnames, f"Missing key for images_data_store: {key}"
        assert "id" not in self.attrnames, "The key 'id' is reserved for this TableStore. Use other key name instead."
        assert (
            "sample" not in self.attrnames
        ), "The key 'sample' is reserved for this TableStore. Use other key name instead."
        self.attrnames_no_filepath = [attrname for attrname in self.attrnames if attrname != "filepath"]
        self.fo_dataset = None
        self.create_dataset_if_empty = create_dataset_if_empty
        # self.semaphore = Semaphore(value=1)

    def get_primary_schema(self) -> DataSchema:
        return self.primary_schema

    def get_meta_schema(self) -> MetaSchema:
        return []

    def _get_dataset(self):
        fo_dataset = None
        datasets = self.fo_session.fiftyone.list_datasets()
        if self.dataset in datasets:
            fo_dataset = self.fo_session.fiftyone.load_dataset(self.dataset)
        return fo_dataset

    def _get_or_create_dataset(self):
        if self.fo_dataset is not None:
            return self.fo_dataset

        # self.semaphore.acquire()
        try:
            datasets = self.fo_session.fiftyone.list_datasets()
            if self.dataset in datasets:
                self.fo_dataset = self.fo_session.fiftyone.load_dataset(self.dataset)
            else:
                if self.create_dataset_if_empty:
                    self.fo_dataset = self.fo_session.fiftyone.Dataset(self.dataset)
                    self.fo_dataset.persistent = True
                else:
                    return ValueError("Dataset {self.dataset} not found.")
        finally:
            # self.semaphore.release()
            pass
        return self.fo_dataset

    def _get_df_sample(self, idx: IndexDF = None) -> pd.DataFrame:
        view = self.fo_session.fiftyone.DatasetView(self._get_or_create_dataset())
        if idx is not None:
            for attrname in self.attrnames:
                if attrname == "filepath":
                    values = idx[attrname].apply(self.mapping_filepath)
                else:
                    values = idx[attrname]

                view = view.select_by(field=attrname, values=values)
        df_sample = pd.DataFrame(
            [
                {
                    "filepath": self.inverse_mapping_filepath(sample.filepath),
                    **{attrname: sample[attrname] for attrname in self.attrnames_no_filepath},
                    "sample": sample,
                }
                for sample in view
            ]
        )
        if len(df_sample) == 0:
            df_sample = pd.DataFrame(columns=self.attrnames + ["sample"])
        if idx is not None:
            df_sample = pd.merge(df_sample, idx, on=self.attrnames_no_filepath)
        return df_sample

    def _attend_image_path_to_image_data(self, image_data: ImageData, idxs_values: List[str]) -> ImageData:
        if self.images_table_store is not None:
            image_path_candidates = self.images_table_store._filenames_from_idxs_values(idxs_values)
            if len(image_path_candidates) > 1:
                for image_path in image_path_candidates:
                    openfile = fsspec.open(image_path)
                    if openfile.fs.exists(openfile.path):
                        break
            else:
                image_path = image_path_candidates[0]
            image_data.image_path = image_path
            for bbox_data in get_all_bboxes_data_in_image_data(image_data):
                bbox_data.image_path = image_path
        assert (
            image_data.image_path is not None and str(image_data.image_path) != ""
        ), f"This {image_data=} have empty image_path"
        return image_data

    def read_rows(self, idx: IndexDF = None, read_data: bool = True) -> DataDF:
        df_sample = self._get_df_sample(idx=idx)
        df_sample["image_data"] = df_sample["sample"].apply(
            lambda sample: (
                self.fo_session.convert_sample_to_image_data(
                    sample=sample,
                    fo_detections_label=self.fo_detections_label,
                    fo_classification_label=self.fo_classification_label,
                    fo_keypoints_label=self.fo_keypoints_label,
                    mapping_filepath=self.inverse_mapping_filepath,
                    additional_info_keys_in_fo_detections=self.additional_info_keys_in_fo_detections,
                    additional_info_keys_in_sample=self.additional_info_keys_in_sample,
                    image_data_cls=self.image_data_cls,
                    bbox_data_cls=self.bbox_data_cls,
                )
                if read_data
                else {}
            )
        )
        return df_sample

    def insert_rows(self, df: DataDF) -> None:
        dataset = self._get_or_create_dataset()

        df_indexes = data_to_index(df, self.attrnames)
        df_current_samples = self.read_rows(df_indexes)
        current_indexes = data_to_index(df_current_samples, self.attrnames)

        idxs_to_be_deleted = index_difference(current_indexes, df_indexes)
        idxs_to_be_added = index_difference(df_indexes, current_indexes)
        idxs_to_be_updated = index_intersection(df_indexes, current_indexes)
        # To be deleted:
        self.delete_rows(idxs_to_be_deleted)

        # To be updated:
        df_to_be_updated = index_to_data(df, idxs_to_be_updated).reset_index(drop=True)
        samples_to_be_updated_unordered = self._get_df_sample(idxs_to_be_updated)["sample"].tolist()
        df_samples_to_be_updated_unordered = pd.DataFrame(
            {
                "sample": [sample for sample in samples_to_be_updated_unordered],
                **{field: [sample[field] for sample in samples_to_be_updated_unordered] for field in self.attrnames},
            }
        )
        samples_to_be_updated = index_to_data(df_samples_to_be_updated_unordered, idxs_to_be_updated)["sample"]
        samples_to_be_updated_with_new_values = [
            self.fo_session.convert_image_data_to_fo_sample(
                image_data=self._attend_image_path_to_image_data(
                    image_data=df_to_be_updated.loc[idx, "image_data"],
                    idxs_values=df_to_be_updated.loc[idx, self.attrnames],
                ),
                fo_detections_label=self.fo_detections_label,
                fo_classification_label=self.fo_classification_label,
                fo_keypoints_label=self.fo_keypoints_label,
                mapping_filepath=self.mapping_filepath,
                additional_info_keys_in_bboxes_data=self.additional_info_keys_in_fo_detections,
                additional_info_keys_in_image_data=self.additional_info_keys_in_sample,
                additional_info={field: df_to_be_updated.loc[idx, field] for field in self.attrnames_no_filepath},
            )
            for idx in df_to_be_updated.index
        ]
        for current_sample, sample_with_updated_values in zip(
            samples_to_be_updated, samples_to_be_updated_with_new_values
        ):
            current_sample.merge(sample_with_updated_values)
        dataset.delete_samples(samples_to_be_updated)  # Работает по поведению так же, как и sample.save()
        dataset.add_samples(samples_to_be_updated, progress=False)

        # To be added:
        df_to_be_added = index_to_data(df, idxs_to_be_added).reset_index(drop=True)
        samples_to_be_added = [
            self.fo_session.convert_image_data_to_fo_sample(
                image_data=self._attend_image_path_to_image_data(
                    image_data=df_to_be_added.loc[idx, "image_data"],
                    idxs_values=df_to_be_added.loc[idx, self.attrnames],
                ),
                fo_detections_label=self.fo_detections_label,
                fo_classification_label=self.fo_classification_label,
                fo_keypoints_label=self.fo_keypoints_label,
                mapping_filepath=self.mapping_filepath,
                additional_info_keys_in_bboxes_data=self.additional_info_keys_in_fo_detections,
                additional_info_keys_in_image_data=self.additional_info_keys_in_sample,
                additional_info={field: df_to_be_added.loc[idx, field] for field in self.attrnames_no_filepath},
            )
            for idx in df_to_be_added.index
        ]
        if len(samples_to_be_added) > 0:
            dataset.add_samples(samples_to_be_added, progress=False)

    def delete_rows(self, idx: IndexDF) -> None:
        dataset = self._get_or_create_dataset()
        df_sample = self._get_df_sample(idx=idx)
        samples = df_sample["sample"].tolist()
        if len(df_sample) > 0:
            if self.rm_only_fo_fields:
                for sample in samples:
                    if sample.has_field(self.fo_detections_label):
                        del sample[self.fo_detections_label]
                    if self.fo_keypoints_label is not None and sample.has_field(self.fo_keypoints_label):
                        del sample[self.fo_keypoints_label]
            dataset.delete_samples(samples)
            if self.rm_only_fo_fields:
                dataset.add_samples(samples, progress=False)



from pathlib import Path
from typing import List, Tuple, Type, Union

from cv_pipeliner.inference_models.classification.core import (
    ClassificationInput,
    ClassificationModel,
    ClassificationModelSpec,
    ClassificationOutput,
)


class Dummy_ClassificationModelSpec(ClassificationModelSpec):
    default_class_name: Union[List[str], str, Path]

    def __post_init__(self):
        self.class_names = [self.default_class_name]

    @property
    def inference_model_cls(self) -> Type["Dummy_ClassificationModel"]:
        from cv_pipeliner.inference_models.classification.dummy import (
            Dummy_ClassificationModel,
        )

        return Dummy_ClassificationModel


class Dummy_ClassificationModel(ClassificationModel):
    def __init__(self, model_spec: Dummy_ClassificationModelSpec, **kwargs):
        assert isinstance(model_spec, Dummy_ClassificationModelSpec)
        super().__init__(model_spec)
        self.default_class_name = model_spec.default_class_name

    def predict(self, input: ClassificationInput, top_n: int = 1) -> ClassificationOutput:
        pred_labels_top_n = [[self.default_class_name for j in range(top_n)] for i in range(len(input))]
        pred_scores_top_n = [[1.0 for j in range(top_n)] for i in range(len(input))]
        return pred_labels_top_n, pred_scores_top_n

    def preprocess_input(self, input: ClassificationInput):
        return input

    @property
    def input_size(self) -> Tuple[int, int]:
        return (None, None)

    @property
    def class_names(self) -> List[str]:
        return [self.default_class_name]



from typing import List

from joblib import Parallel, delayed

from cv_pipeliner.core.batch_generator import BatchGenerator
from cv_pipeliner.core.data import ImageData


class BatchGeneratorImageData(BatchGenerator):
    def __init__(
        self,
        data: List[ImageData],
        batch_size: int,
        use_not_caught_elements_as_last_batch: bool = True,
        max_workers: int = 64,
    ):
        self.max_workers = max_workers
        assert all(isinstance(d, ImageData) for d in data)
        super().__init__(data, batch_size, use_not_caught_elements_as_last_batch)

    def __getitem__(self, index) -> List[ImageData]:
        batch = super().__getitem__(index)
        Parallel(n_jobs=self.max_workers, prefer="threads")(
            delayed(lambda x: x.open_image(inplace=True))(image_data) for image_data in batch
        )
        return batch

    def __iter__(self):
        for images_data in (self[i] for i in range(len(self))):
            yield images_data
            for image_data in images_data:
                image_data.image = None
            del images_data



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.utils.images import rotate_point
from cv_pipeliner.utils.images_datas import combine_mask_polygons_to_one_polygon


def parse_rectangle_labels_to_bbox_data(rectangle_label: Dict) -> BboxData:
    original_height = rectangle_label["original_height"]
    original_width = rectangle_label["original_width"]
    height = rectangle_label["value"]["height"]
    width = rectangle_label["value"]["width"]
    xmin = rectangle_label["value"]["x"]
    ymin = rectangle_label["value"]["y"]
    label = rectangle_label["value"]["rectanglelabels"][0]
    xmax = xmin + width
    ymax = ymin + height
    xmin = max(0, min(original_width - 1, xmin / 100 * original_width))
    ymin = max(0, min(original_height - 1, ymin / 100 * original_height))
    xmax = max(0, min(original_width - 1, xmax / 100 * original_width))
    ymax = max(0, min(original_height - 1, ymax / 100 * original_height))
    angle = rectangle_label["value"]["rotation"]
    points = [(xmin, ymin), (xmin, ymax), (xmax, ymin), (xmax, ymax)]
    rotated_points = [rotate_point(x=x, y=y, cx=xmin, cy=ymin, angle=angle) for (x, y) in points]
    xmin = max(0, min([x for (x, y) in rotated_points]))
    ymin = max(0, min([y for (x, y) in rotated_points]))
    xmax = max([x for (x, y) in rotated_points])
    ymax = max([y for (x, y) in rotated_points])
    xmin = max(0, min(original_width - 1, xmin))
    ymin = max(0, min(original_height - 1, ymin))
    xmax = max(0, min(original_width - 1, xmax))
    ymax = max(0, min(original_height - 1, ymax))
    bbox_data = BboxData(
        xmin=round(xmin),
        ymin=round(ymin),
        xmax=round(xmax),
        ymax=round(ymax),
        label=label,
        meta_height=original_height,
        meta_width=original_width,
    )
    return bbox_data


def convert_image_data_to_rectangle_labels(
    image_data: ImageData,
    from_name: str,
    to_name: str,
) -> Dict:
    im_width, im_height = image_data.get_image_size(exif_transpose=False)
    rectangle_labels = []
    for bbox_data in image_data.bboxes_data:
        rectangle_label = {
            "original_width": im_width,
            "original_height": im_height,
            "image_rotation": 0,
            "value": {
                "x": bbox_data.xmin / im_width * 100,
                "y": bbox_data.ymin / im_height * 100,
                "width": (bbox_data.xmax - bbox_data.xmin) / im_width * 100,
                "height": (bbox_data.ymax - bbox_data.ymin) / im_height * 100,
                "rotation": 0,
                "rectanglelabels": [bbox_data.label],
            },
            "from_name": from_name,
            "to_name": to_name,
            "type": "rectanglelabels",
        }
        rectangle_labels.append(rectangle_label)
    return rectangle_labels


def parse_polygon_label_to_bbox_data(
    polygon_label: Dict,
    keypoints_from_name: Optional[str],
    mask_from_name: Optional[str],
) -> BboxData:
    original_height = polygon_label["original_height"]
    original_width = polygon_label["original_width"]
    keypoints = []
    for x, y in polygon_label["value"]["points"]:
        x = x / 100 * polygon_label["original_width"]
        y = y / 100 * polygon_label["original_height"]
        keypoints.append([round(max(0, min(original_width - 1, x))), round(max(0, min(original_height - 1, y)))])
    keypoints = np.array(keypoints)
    bbox_data = BboxData(
        xmin=round(np.min(keypoints[:, 0])),
        ymin=round(np.min(keypoints[:, 1])),
        xmax=round(np.max(keypoints[:, 0])),
        ymax=round(np.max(keypoints[:, 1])),
        keypoints=(np.round(keypoints) if polygon_label["from_name"] == keypoints_from_name else None),
        mask=([keypoints] if polygon_label["from_name"] == mask_from_name else None),
        label=polygon_label["value"]["polygonlabels"][0],
    )
    return bbox_data


def parse_polygon_label_to_mask(polygon_label: Dict):
    original_height = polygon_label["original_height"]
    original_width = polygon_label["original_width"]
    mask = []
    for x, y in polygon_label["value"]["points"]:
        x = x / 100 * polygon_label["original_width"]
        y = y / 100 * polygon_label["original_height"]
        mask.append([round(max(0, min(original_width - 1, x))), round(max(0, min(original_height - 1, y)))])
    labels = polygon_label["value"]["polygonlabels"]
    label = labels[0] if len(labels) == 1 else None
    return [mask], label


def convert_image_data_to_polygon_label(
    image_data: ImageData,
    from_name: str,
    to_name: str,
    polygonlabels: str,
) -> Dict:
    im_width, im_height = image_data.get_image_size(exif_transpose=False)
    rectangle_labels = []
    for bbox_data in image_data.bboxes_data:
        rectangle_labels.append(
            {
                "original_width": im_width,
                "original_height": im_height,
                "image_rotation": 0,
                "value": {
                    "points": [[x * 100 / im_width, y * 100 / im_height] for x, y in bbox_data.keypoints],
                    "polygonlabels": [polygonlabels],
                },
                "from_name": from_name,
                "to_name": to_name,
                "type": "polygonlabels",
            }
        )
    return rectangle_labels


def convert_image_data_to_keypoint_label(
    image_data: ImageData, from_name: str, to_name: str, keypoints_labels: List[str], keypoints_width: float = 0.8
) -> Dict:
    im_width, im_height = image_data.get_image_size(exif_transpose=False)
    keypoints_labels_json = []
    keypoints_to_be_added = [image_data.keypoints] + [bbox_data.keypoints for bbox_data in image_data.bboxes_data]
    for keypoints in keypoints_to_be_added:
        if len(keypoints) == 0:
            continue
        assert len(keypoints) == len(
            keypoints_labels
        ), f"KeypointsLabels  lengthmismatch: {keypoints=}, {keypoints_labels=}"
        for keypoint, keypointlabel in zip(keypoints, keypoints_labels):
            x, y = keypoint[0], keypoint[1]
            keypoints_labels_json.append(
                {
                    "original_width": im_width,
                    "original_height": im_height,
                    "image_rotation": 0,
                    "value": {
                        "x": x * 100 / im_width,
                        "y": y * 100 / im_height,
                        "width": keypoints_width,
                        "keypointlabels": [keypointlabel],
                    },
                    "from_name": from_name,
                    "to_name": to_name,
                    "type": "keypointlabels",
                }
            )
    return keypoints_labels_json


def parse_keypoint_label_to_keypoint(keypoint_label: Dict) -> list:
    original_height = keypoint_label["original_height"]
    original_width = keypoint_label["original_width"]
    x, y = keypoint_label["value"]["x"], keypoint_label["value"]["y"]
    x = x / 100 * original_width
    y = y / 100 * original_height
    x = round(x)
    y = round(y)
    labels = keypoint_label["value"]["keypointlabels"]
    label = labels[0] if len(labels) == 1 else None
    return [x, y], label


def convert_image_data_to_annotation(
    image_data: ImageData,
    to_name: str,
    bboxes_from_name: Optional[str] = None,
    label_from_name: Optional[str] = None,
    keypoints_from_name: Optional[str] = None,
    keypoints_labels: Optional[List[str]] = None,
    mask_from_name: Optional[str] = None,
    keypoints_width: float = 0.8,
) -> List[Dict[str, Any]]:
    # При импорте разметки в LS не учитываются Exif теги, но при экспорте оно учитывается
    im_width, im_height = image_data.get_image_size(exif_transpose=False)
    annotations = []
    if label_from_name is not None and image_data.label is not None:
        annotations.append(
            {
                "value": {"choices": [image_data.label]},
                "from_name": label_from_name,
                "to_name": to_name,
                "type": "choices",
            }
        )

    if bboxes_from_name is not None:
        for bbox_idx, bbox_data in enumerate(image_data.bboxes_data):
            annotations.append(
                {
                    "id": f"bbox{bbox_idx}",
                    "original_width": im_width,
                    "original_height": im_height,
                    "image_rotation": 0,
                    "value": {
                        "x": bbox_data.xmin / im_width * 100,
                        "y": bbox_data.ymin / im_height * 100,
                        "width": (bbox_data.xmax - bbox_data.xmin) / im_width * 100,
                        "height": (bbox_data.ymax - bbox_data.ymin) / im_height * 100,
                        "rotation": 0,
                        "rectanglelabels": [bbox_data.label],
                    },
                    "from_name": bboxes_from_name,
                    "to_name": to_name,
                    "type": "rectanglelabels",
                }
            )
            if keypoints_from_name is not None:
                if len(bbox_data.keypoints) == 0:
                    continue
                for kp_idx, keypoint in enumerate(bbox_data.keypoints):
                    if keypoints_labels is not None:
                        assert len(keypoints_labels) == len(bbox_data.keypoints)
                    x, y = keypoint[0], keypoint[1]
                    annotations.append(
                        {
                            "id": f"bbox{bbox_idx}_kp{kp_idx}",
                            "original_width": im_width,
                            "original_height": im_height,
                            "image_rotation": 0,
                            "value": {
                                "x": x * 100 / im_width,
                                "y": y * 100 / im_height,
                                "width": keypoints_width,
                                "keypointlabels": ([keypoints_labels[kp_idx]] if keypoints_labels is not None else []),
                            },
                            "from_name": keypoints_from_name,
                            "to_name": to_name,
                            "type": "keypointlabels",
                        }
                    )
                    annotations.append(
                        {
                            "from_id": f"bbox{bbox_idx}_kp{kp_idx}",
                            "to_id": f"bbox{bbox_idx}",
                            "type": "relation",
                            "direction": "bi",
                        }
                    )
            if mask_from_name is not None:
                assert len(bbox_data.mask) == 1
                annotations.append(
                    {
                        "id": f"bbox{bbox_idx}_mask",
                        "original_width": im_width,
                        "original_height": im_height,
                        "image_rotation": 0,
                        "value": {
                            "points": [
                                [x * 100 / im_width, y * 100 / im_height]
                                for x, y in combine_mask_polygons_to_one_polygon(bbox_data.mask)
                            ],
                            "polygonlabels": [bbox_data.label],
                        },
                        "from_name": mask_from_name,
                        "to_name": to_name,
                        "type": "polygonlabels",
                    }
                )
                annotations.append(
                    {
                        "from_id": f"bbox{bbox_idx}_mask",
                        "to_id": f"bbox{bbox_idx}",
                        "type": "relation",
                        "direction": "bi",
                    }
                )

    if len(image_data.keypoints) > 0 and keypoints_from_name is not None:
        if keypoints_labels is not None:
            assert len(image_data.keypoints) == len(
                keypoints_labels
            ), f"KeypointsLabels mismatch: {image_data.keypoints=}, {keypoints_labels=}"
        for kp_idx, (keypoint, keypointlabel) in enumerate(zip(image_data.keypoints, keypoints_labels)):
            x, y = keypoint[0], keypoint[1]
            annotations.append(
                {
                    "id": f"kp{kp_idx}",
                    "original_width": im_width,
                    "original_height": im_height,
                    "image_rotation": 0,
                    "value": {
                        "x": x * 100 / im_width,
                        "y": y * 100 / im_height,
                        "width": keypoints_width,
                        "keypointlabels": [keypointlabel],
                    },
                    "from_name": keypoints_from_name,
                    "to_name": to_name,
                    "type": "keypointlabels",
                }
            )

    if len(image_data.mask) > 0 and mask_from_name is not None:
        annotations.append(
            {
                "id": "mask",
                "original_width": im_width,
                "original_height": im_height,
                "image_rotation": 0,
                "value": {
                    "points": [
                        [x * 100 / im_width, y * 100 / im_height]
                        for x, y in combine_mask_polygons_to_one_polygon(image_data.mask)
                    ],
                    "polygonlabels": [image_data.label],
                },
                "from_name": mask_from_name,
                "to_name": to_name,
                "type": "polygonlabels",
            }
        )

    return annotations


def parse_result(
    result: Dict[str, Any],
    bboxes_from_name: Optional[str],
    label_from_name: Optional[str],
    keypoints_from_name: Optional[str],
    mask_from_name: Optional[str],
) -> Optional[Union[str, Tuple[str, Dict[str, Any], str], Tuple[str, Tuple[float, float], str, str]]]:
    from_name = result.get("from_name", None)
    if from_name is None:
        return None
    if label_from_name and result["from_name"] == label_from_name:
        if len(result["value"]["choices"]) > 0:
            return result["value"]["choices"][0]
    if bboxes_from_name and result["from_name"] == bboxes_from_name:
        if result["type"] == "rectanglelabels":
            return "bbox", parse_rectangle_labels_to_bbox_data(result), result["id"]
        elif result["type"] == "polygonlabels":
            return "bbox", parse_polygon_label_to_bbox_data(result, keypoints_from_name, mask_from_name), result["id"]
    if keypoints_from_name:
        if result["type"] == "keypointlabels" and result["from_name"] == keypoints_from_name:
            (x, y), kp_label = parse_keypoint_label_to_keypoint(result)
            return "keypoint", (x, y), kp_label, result["id"]
    if mask_from_name:
        if result["type"] == "polygonlabels" and result["from_name"] == mask_from_name:
            mask, mask_label = parse_polygon_label_to_mask(result)
            if result["from_name"] == mask_from_name:
                return "mask", mask, mask_label, result["id"]
    return None


def handle_relations(
    result: Dict[str, Any],
    id_to_keypoint_idx: Dict[str, int],
    id_to_mask_idx: Dict[str, int],
    bbox_id_to_keypoints_idxs_relation: Dict[str, List[str]],
    keypoints_idxs_that_have_relation: set,
    bbox_id_to_masks_idxs_relation: Dict[str, List[str]],
    masks_idxs_that_have_relation: set,
) -> None:
    keypoint_idx: Optional[str] = None
    mask_idx: Optional[str] = None
    if result["from_id"] in id_to_keypoint_idx:
        keypoint_idx = result["from_id"]
        bbox_idx = result["to_id"]
    elif result["to_id"] in id_to_keypoint_idx:
        keypoint_idx = result["to_id"]
        bbox_idx = result["from_id"]
    if result["from_id"] in id_to_mask_idx:
        mask_idx = result["from_id"]
        bbox_idx = result["to_id"]
    elif result["to_id"] in id_to_mask_idx:
        mask_idx = result["to_id"]
        bbox_idx = result["from_id"]
    if keypoint_idx is not None:
        bbox_id_to_keypoints_idxs_relation.setdefault(bbox_idx, []).append(keypoint_idx)
        keypoints_idxs_that_have_relation.add(keypoint_idx)
    if mask_idx is not None:
        bbox_id_to_masks_idxs_relation.setdefault(bbox_idx, []).append(mask_idx)
        masks_idxs_that_have_relation.add(mask_idx)


def process_annotations(
    annotation: List[Dict[str, Any]],
    bboxes_from_name: Optional[str],
    label_from_name: Optional[str],
    keypoints_from_name: Optional[str],
    mask_from_name: Optional[str],
) -> Tuple[
    List[Dict[str, Any]],
    List[Tuple[float, float]],
    List[Tuple[float, float]],
    List[str],
    List[str],
    Optional[str],
]:
    bboxes_data: List[Dict[str, Any]] = []
    id_to_bbox_data_idx: Dict[str, int] = {}
    keypoints: List[Tuple[float, float]] = []
    masks: List[Tuple[float, float]] = []
    items_keypoints_labels: List[str] = []
    items_mask_labels: List[str] = []
    id_to_keypoint_idx: Dict[str, int] = {}
    id_to_mask_idx: Dict[str, int] = {}
    bbox_id_to_keypoints_idxs_relation: Dict[str, List[str]] = {}
    bbox_id_to_masks_idxs_relation: Dict[str, List[str]] = {}
    keypoints_idxs_that_have_relation: set = set()
    masks_idxs_that_have_relation: set = set()
    image_data_label: Optional[str] = None

    for result in annotation["result"]:
        parsed_result = parse_result(result, bboxes_from_name, label_from_name, keypoints_from_name, mask_from_name)
        if parsed_result:
            if isinstance(parsed_result, str):
                if image_data_label:
                    raise ValueError(f"Found duplicated choices: {result=} (previously choice is {image_data_label=})")
                image_data_label = parsed_result
            elif parsed_result[0] == "bbox":
                bboxes_data.append(parsed_result[1])
                id_to_bbox_data_idx[parsed_result[2]] = len(bboxes_data) - 1
            elif parsed_result[0] == "keypoint":
                keypoints.append(parsed_result[1])
                items_keypoints_labels.append(parsed_result[2])
                id_to_keypoint_idx[parsed_result[3]] = len(keypoints) - 1
            elif parsed_result[0] == "mask":
                masks.append(parsed_result[1])
                items_mask_labels.append(parsed_result[2])
                id_to_mask_idx[parsed_result[3]] = len(masks) - 1

        if result["type"] == "relation":
            handle_relations(
                result,
                id_to_keypoint_idx,
                id_to_mask_idx,
                bbox_id_to_keypoints_idxs_relation,
                keypoints_idxs_that_have_relation,
                bbox_id_to_masks_idxs_relation,
                masks_idxs_that_have_relation,
            )

    return (
        bboxes_data,
        keypoints,
        masks,
        items_keypoints_labels,
        items_mask_labels,
        image_data_label,
    )


def sort_items_by_labels(
    items: List[Union[Tuple[float, float], Tuple[float, float]]], labels: List[str], label_to_position: Dict[str, int]
) -> np.ndarray:
    item_labels_positions = list(map(label_to_position.get, labels))
    item_sorted_idxs = np.argsort(item_labels_positions)
    return np.array(items)[item_sorted_idxs]


def convert_annotation_to_image_data(
    annotation: List[Dict[str, Any]],
    bboxes_from_name: Optional[str] = None,
    label_from_name: Optional[str] = None,
    keypoints_from_name: Optional[str] = None,
    keypoints_labels: Optional[List[str]] = None,
    mask_from_name: Optional[str] = None,
    mask_labels: Optional[List[str]] = None,
    image_path: Optional[str] = None,
) -> ImageData:
    (
        bboxes_data,
        keypoints,
        masks,
        items_keypoints_labels,
        items_mask_labels,
        image_data_label,
    ) = process_annotations(annotation, bboxes_from_name, label_from_name, keypoints_from_name, mask_from_name)

    if keypoints_labels is not None:
        label_to_position = {label: idx for idx, label in enumerate(keypoints_labels)}
        keypoints = sort_items_by_labels(keypoints, items_keypoints_labels, label_to_position)

    if mask_labels is not None:
        label_to_position = {label: idx for idx, label in enumerate(mask_labels)}
        masks = sort_items_by_labels(masks, items_mask_labels, label_to_position)

    return ImageData(
        image_path=image_path, bboxes_data=bboxes_data, label=image_data_label, keypoints=keypoints, mask=masks
    )



import base64
import io
import json
import math
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

import cv2
import fsspec
import imageio
import imutils
import numpy as np
from matplotlib.figure import Figure
from pathy import Pathy
from PIL import Image, ImageDraw, ImageFont
from tqdm import tqdm

from cv_pipeliner.logging import logger

if hasattr(Image, "Transpose"):
    TransposeModule = Image.Transpose
else:
    TransposeModule = Image

EXIF_ORIENTATION_TO_METHOD = {
    2: TransposeModule.FLIP_LEFT_RIGHT,
    3: TransposeModule.ROTATE_180,
    4: TransposeModule.FLIP_TOP_BOTTOM,
    5: TransposeModule.TRANSPOSE,
    6: TransposeModule.ROTATE_270,
    7: TransposeModule.TRANSVERSE,
    8: TransposeModule.ROTATE_90,
}


def exif_transpose_image(image: Image.Image) -> Tuple[int, int]:
    """
    If an image has an EXIF Orientation tag, return a size of image that is
    transposed accordingly. Otherwise, return the size of image itself.
    """
    exif = image.getexif()
    orientation = exif.get(0x0112)
    method = EXIF_ORIENTATION_TO_METHOD.get(orientation)

    if method is not None:
        return image.transpose(method)

    return image


def denormalize_bboxes(
    bboxes: List[Tuple[float, float, float, float]], image_width: int, image_height: int
) -> List[Tuple[int, int, int, int]]:
    """
    Denormalize normalized bboxes coordinates.
    bboxes must have this format: (xmin, ymin, xmax, ymax)
    """
    bboxes = np.array(bboxes.copy())
    bboxes[:, [0, 2]] = bboxes[:, [0, 2]] * image_width
    bboxes[:, [1, 3]] = bboxes[:, [1, 3]] * image_height
    bboxes = bboxes.round().astype(int)
    bboxes[:, [0, 2]] = np.clip(bboxes[:, [0, 2]], 0, image_width - 1)
    bboxes[:, [1, 3]] = np.clip(bboxes[:, [1, 3]], 0, image_height - 1)
    return bboxes


def rescale_bboxes_with_pad(
    bboxes: List[Tuple[int, int, int, int]],  # (xmins, ymins, xmaxs, ymaxs)
    current_width: int,
    current_height: int,
    target_width: int,
    target_height: int,
) -> List[Tuple[int, int, int, int]]:
    # Taken from https://github.com/ultralytics/yolov5/blob/master/utils/general.py
    # Rescale coords (xyxy) from target_tensor to img0_shape
    gain = min(current_width / target_width, current_height / target_height)
    pad_w = (current_width - target_width * gain) / 2
    pad_h = (current_height - target_height * gain) / 2  # wh padding
    bboxes = np.array(bboxes, dtype=np.float32).reshape(-1, 4)
    bboxes[:, [0, 2]] -= pad_w
    bboxes[:, [1, 3]] -= pad_h
    bboxes /= gain
    bboxes = bboxes.round().astype(int)
    bboxes[:, [0, 2]] = np.clip(bboxes[:, [0, 2]], 0, target_width - 1)
    bboxes[:, [1, 3]] = np.clip(bboxes[:, [1, 3]], 0, target_height - 1)
    return bboxes


def cut_bboxes_from_image(image: np.ndarray, bboxes: List[Tuple[int, int, int, int]]) -> List[np.ndarray]:
    img_boxes = []
    for bbox in bboxes:
        xmin, ymin, xmax, ymax = bbox
        img_boxes.append(image[ymin:ymax, xmin:xmax])
    return img_boxes


def get_img_from_fig(fig: Figure) -> np.ndarray:
    buf = io.BytesIO()
    fig.savefig(buf, format="png")
    buf.seek(0)
    img_arr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
    buf.close()
    img = cv2.imdecode(img_arr, 1)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img


def rotate_point(x: float, y: float, cx: float, cy: float, angle: float) -> Tuple[float, float]:
    angle = math.radians(angle)
    xnew = cx + (x - cx) * math.cos(angle) - (y - cy) * math.sin(angle)
    ynew = cy + (x - cx) * math.sin(angle) + (y - cy) * math.cos(angle)
    return xnew, ynew


def is_base64(s: str):
    try:
        return base64.b64encode(base64.b64decode(s)).decode() == s
    except Exception:
        return False


def open_image(
    image: Union[str, Path, fsspec.core.OpenFile, bytes, io.BytesIO, Image.Image],
    open_as_rgb: bool = False,
    exif_transpose: bool = False,
) -> np.ndarray:
    if isinstance(image, str) or isinstance(image, Path) or isinstance(image, Pathy):
        if is_base64(str(image)):
            image_bytes = base64.b64decode(str(image))
        else:
            with fsspec.open(str(image), "rb") as src:
                image_bytes = src.read()
    elif isinstance(image, fsspec.core.OpenFile):
        with image as src:
            image_bytes = src.read()
    elif isinstance(image, bytes):
        image_bytes = image
    elif isinstance(image, io.BytesIO):
        image_bytes = image.getvalue()
    elif isinstance(image, Image.Image):
        image_bytes = np.array(image)
    else:
        raise ValueError(f"Got unknown type: {type(image)}.")
    if not isinstance(image_bytes, np.ndarray):
        try:
            image = np.array(imageio.v3.imread(image_bytes, rotate=exif_transpose))
        except KeyError:  # there is bug with rotation in imageio
            image = np.array(imageio.v3.imread(image_bytes))
    else:
        image = image_bytes
    if open_as_rgb:
        if image.shape[-1] == 4:
            image = cv2.cvtColor(image, cv2.COLOR_RGBA2RGB)
        if len(image.shape) == 2 or image.shape[-1] == 1:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)

    return image


def concat_images(
    image_a: np.ndarray,
    image_b: np.ndarray,
    background_color_a: Tuple[int, int, int, int] = None,
    background_color_b: Tuple[int, int, int, int] = None,
    thumbnail_size_a: Tuple[int, int] = None,
    thumbnail_size_b: Tuple[int, int] = None,
    how: Literal["horizontally", "vertically"] = "horizontally",
    mode: Literal["L", "RGB", "RGBA"] = "RGBA",
    background_edge_width: int = 3,
    between_edge_width: int = 0,
) -> np.ndarray:
    if image_a is None and image_b is not None:
        return image_b
    if image_a is not None and image_b is None:
        return image_a
    if len(image_a.shape) == 2 or image_a.shape[-1] == 1:
        image_a = cv2.cvtColor(image_a, cv2.COLOR_GRAY2RGBA)
    elif image_a.shape[-1] == 3:
        image_a = cv2.cvtColor(image_a, cv2.COLOR_RGB2RGBA)
    if len(image_b.shape) == 2 or image_b.shape[-1] == 1:
        image_b = cv2.cvtColor(image_b, cv2.COLOR_GRAY2RGBA)
    elif image_b.shape[-1] == 3:
        image_b = cv2.cvtColor(image_b, cv2.COLOR_RGB2RGBA)
    if thumbnail_size_a is not None:
        image_a = Image.fromarray(image_a)
        image_a.thumbnail(thumbnail_size_b)
        image_a = np.array(image_a)
    if thumbnail_size_b is not None:
        image_b = Image.fromarray(image_b)
        image_b.thumbnail(thumbnail_size_b)
        image_b = np.array(image_b)

    ha, wa = image_a.shape[:2]
    hb, wb = image_b.shape[:2]

    if how == "horizontally":
        max_height = np.max([ha, hb])
        total_width = wa + wb + between_edge_width

        min_ha = max_height // 2 - ha // 2
        max_ha = max_height // 2 + ha // 2
        min_hb = max_height // 2 - hb // 2
        max_hb = max_height // 2 + hb // 2

        new_image = np.zeros(shape=(max_height, total_width, 4), dtype=np.uint8)
        new_image[min_ha:max_ha, :wa, :] = image_a[0 : (max_ha - min_ha), :]
        new_image[min_hb:max_hb, wa + between_edge_width : wa + between_edge_width + wb, :] = image_b[
            0 : (max_hb - min_hb), :
        ]

        if background_color_a is not None:
            new_image[:background_edge_width, :wa, :] = background_color_a
            new_image[-background_edge_width:, :wa, :] = background_color_a
            new_image[:, :background_edge_width, :] = background_color_a
            new_image[:, wa + between_edge_width - (background_edge_width - 1) : wa + between_edge_width, :] = (
                background_color_a
            )
        if background_color_b is not None:
            new_image[:background_edge_width, wa:, :] = background_color_b
            new_image[-background_edge_width:, wa:, :] = background_color_b
            new_image[:, -background_edge_width:, :] = background_color_b
            new_image[:, wa + between_edge_width : wa + between_edge_width + (background_edge_width - 1), :] = (
                background_color_b
            )
    elif how == "vertically":
        max_width = np.max([wa, wb])
        total_height = ha + hb + between_edge_width

        min_wa = max_width // 2 - wa // 2
        max_wa = max_width // 2 + wa // 2
        min_wb = max_width // 2 - wb // 2
        max_wb = max_width // 2 + wb // 2

        new_image = np.zeros(shape=(total_height, max_width, 4), dtype=np.uint8)
        new_image[:ha, min_wa:max_wa, :] = image_a[:, 0 : (max_wa - min_wa)]
        new_image[ha + between_edge_width : ha + between_edge_width + hb, min_wb:max_wb, :] = image_b[
            :, 0 : (max_wb - min_wb)
        ]

        if background_color_a is not None:
            new_image[:ha, :background_edge_width, :] = background_color_a
            new_image[:ha, -background_edge_width:, :] = background_color_a
            new_image[:background_edge_width, :, :] = background_color_a
            new_image[ha + between_edge_width - (background_edge_width - 1) : ha + between_edge_width, :, :] = (
                background_color_a
            )
        if background_color_b is not None:
            new_image[ha:, :background_edge_width, :] = background_color_b
            new_image[ha:, -background_edge_width:, :] = background_color_b
            new_image[-background_edge_width:, :, :] = background_color_b
            new_image[ha + between_edge_width : ha + between_edge_width + (background_edge_width - 1), :, :] = (
                background_color_b
            )
    else:
        raise ValueError("Parametr how must be 'horizontally' or 'vertically'")

    if mode == "RGBA":
        pass
    elif mode == "RGB":
        new_image = cv2.cvtColor(new_image, cv2.COLOR_RGBA2RGB)
    elif mode == "L":
        new_image = cv2.cvtColor(new_image, cv2.COLOR_RGBA2GRAY)
    else:
        raise ValueError("Parametr mode must be 'RGBA' or 'RGB' or 'L'")

    return new_image


def put_text_on_image(
    image: np.ndarray,
    text: str,
    fontsize: int,
    ymax: int,
    maximum_width: int,
    font: str = "arial.ttf",
    fill: str = "black",
) -> np.ndarray:
    texts = [text[i : i + maximum_width] for i in range(0, len(text), maximum_width)]
    image_pil = Image.fromarray(image)
    draw = ImageDraw.Draw(image_pil)
    try:
        font = ImageFont.truetype(font, fontsize)
    except IOError:
        font = ImageFont.load_default()
    width, height = image_pil.size
    for i, subtext in enumerate(texts):
        text_width, text_height = font.getsize(subtext)
        left = (width - text_width) // 2
        text_bottom = ymax + (i + 1) * fontsize
        margin = np.ceil(0.05 * text_height)
        draw.text(xy=(left + margin, text_bottom - text_height - margin), text=subtext, fill=fill, font=font)
    image = np.array(image_pil)
    return image


def draw_rectangle(
    image: np.ndarray,
    xmin: int,
    ymin: int,
    xmax: int,
    ymax: int,
    color: Tuple[int, int, int],
    thickness: int,
    alpha: float,
):
    image_original = image.copy()
    image = image.copy()

    image[max(0, ymin - thickness) : ymin, max(0, xmin - thickness) : (xmax + thickness), :] = np.uint8(
        (1 - alpha) * image_original[max(0, ymin - thickness) : ymin, max(0, xmin - thickness) : (xmax + thickness), :]
        + alpha * np.array(color)  # noqa: E501
    )
    image[ymax : ymax + thickness, max(0, xmin - thickness) : (xmax + thickness), :] = np.uint8(
        (1 - alpha) * image_original[ymax : ymax + thickness, max(0, xmin - thickness) : (xmax + thickness), :]
        + alpha * np.array(color)  # noqa: E501
    )
    image[max(0, ymin - thickness) : (ymax + thickness), xmax : xmax + thickness, :] = np.uint8(
        (1 - alpha) * image_original[max(0, ymin - thickness) : (ymax + thickness), xmax : xmax + thickness, :]
        + alpha * np.array(color)  # noqa: E501
    )
    image[max(0, ymin - thickness) : ymax + thickness, max(0, xmin - thickness) : xmin, :] = np.uint8(
        (1 - alpha) * image_original[max(0, ymin - thickness) : ymax + thickness, max(0, xmin - thickness) : xmin, :]
        + alpha * np.array(color)  # noqa: E501
    )
    return image


# TODO: Remove 4D padding as it's very slow.
def get_base_label_image_with_description(
    base_label_image: np.ndarray,
    label: str,
    description: str,
    pad_color: Tuple[int, int, int, int] = (255, 255, 255, 255),
    pad_resize: int = 90,
    pad_width: int = 5,
    base_resize: int = 150,
    maximum_text_width: int = 20,
    fill: str = "black",
) -> np.ndarray:
    if len(base_label_image.shape) == 2 or base_label_image.shape[-1] == 1:
        base_label_image = cv2.cvtColor(base_label_image, cv2.COLOR_GRAY2RGBA)
    elif base_label_image.shape[-1] == 3:
        base_label_image = cv2.cvtColor(base_label_image, cv2.COLOR_RGB2RGBA)
    base_label_image = imutils.resize(base_label_image, width=base_resize, height=base_resize)
    height, width, _ = base_label_image.shape
    base_label_image = np.pad(
        base_label_image,
        pad_width=(
            (max(0, pad_resize - height // 2), max(0, pad_resize - height // 2)),
            (max(0, pad_resize - width // 2), max(0, pad_resize - width // 2)),
            (0, 0),
        ),
        constant_values=((pad_color, pad_color), (pad_color, pad_color), (0, 0)),
        mode="constant",
    )
    how_many = 30 * len(description) // maximum_text_width
    base_label_image = np.pad(
        base_label_image,
        pad_width=((how_many, 0), (0, 0), (0, 0)),
        constant_values=((pad_color, 0), (0, 0), (0, 0)),
        mode="constant",
    )
    base_label_image = np.pad(
        base_label_image,
        pad_width=((pad_width, pad_width), (pad_width, pad_width), (0, 0)),
        constant_values=((pad_color, pad_color), (pad_color, pad_color), (0, 0)),
        mode="constant",
    )
    base_label_image = np.pad(
        base_label_image,
        pad_width=((60, 0), (0, 0), (0, 0)),
        constant_values=((pad_color, 0), (0, 0), (0, 0)),
        mode="constant",
    )
    fontsize1, fontsize2 = 30, 17
    ymax1, ymax2 = 25, 10
    base_label_image = put_text_on_image(
        image=base_label_image, text=label, fontsize=fontsize1, ymax=ymax1, maximum_width=maximum_text_width, fill=fill
    )
    base_label_image = put_text_on_image(
        image=base_label_image,
        text=description,
        fontsize=fontsize2,
        ymax=fontsize1 + ymax1 + ymax2,
        maximum_width=maximum_text_width,
        fill=fill,
    )

    return base_label_image


def get_label_to_description(
    label_to_description_dict: Union[str, Path, Dict, None], default_description: str = "No description."
) -> Dict[str, str]:
    if label_to_description_dict is None:
        label_to_description_dict = {}
    elif isinstance(label_to_description_dict, str) or isinstance(label_to_description_dict, Path):
        with fsspec.open(label_to_description_dict, "r") as src:
            label_to_description_dict = json.load(src)

    label_to_description = defaultdict(lambda: default_description)
    label_to_description["unknown"] = default_description
    for k in label_to_description_dict:
        label_to_description[k] = label_to_description_dict[k]

    return label_to_description


def get_label_to_base_label_image(
    base_labels_images: Union[str, Path],
    label_to_description: Union[str, Path, Dict[str, str]] = None,
    add_label_to_image: bool = False,
    make_labels_for_these_class_names_too: List[str] = [],  # add known description to classes without base images
) -> Dict[str, np.ndarray]:
    if base_labels_images is None:
        return None

    base_labels_images_files = fsspec.open_files(str(base_labels_images))
    ann_class_names_files = [
        Pathy.fluid(base_label_image_file.path).stem for base_label_image_file in base_labels_images_files
    ]
    unique_ann_class_names = set(ann_class_names_files)
    if "unknown" not in unique_ann_class_names:
        raise ValueError(f'"{base_labels_images}" must have image with name "unknown.*"')
    unknown_image_path = base_labels_images_files[ann_class_names_files.index("unknown")]
    label_to_base_label_image = defaultdict(lambda: unknown_image_path)
    label_to_base_label_image["unknown"] = unknown_image_path
    logger.info(f"Loading base labels images from {base_labels_images}...")
    for label in tqdm(list(unique_ann_class_names) + list(set(make_labels_for_these_class_names_too))):
        if label in unique_ann_class_names:
            base_label_image = base_labels_images_files[ann_class_names_files.index(label)]
        else:
            base_label_image = label_to_base_label_image["unknown"]
        label_to_base_label_image[label] = base_label_image

    def label_to_base_label_image_func(
        label: str,
        label_to_description: Union[str, Path, Dict[str, str]] = label_to_description,
        add_label_to_image: bool = add_label_to_image,
    ):
        base_label_image = open_image(label_to_base_label_image[label])
        if label_to_description is not None:
            if isinstance(label_to_description, str) or isinstance(label_to_description, Path):
                label_to_description = get_label_to_description(label_to_description_dict=label_to_description)
            base_label_image = get_base_label_image_with_description(
                base_label_image=base_label_image, label=label, description=label_to_description[label]
            )
        elif add_label_to_image:
            base_label_image = get_base_label_image_with_description(
                base_label_image=base_label_image, label=label, description=""
            )
        return base_label_image

    return label_to_base_label_image_func


def draw_n_base_labels_images(
    labels: List[str],
    label_to_base_label_image: Dict[str, np.ndarray],
    label_to_description: Union[str, Path, Dict[str, str]] = None,
    images_per_row: int = 7,
) -> np.ndarray:
    rows = np.array_split(labels, max(1, len(labels) // images_per_row))
    total_image = None
    i = 0
    for row in rows:
        total_image_row = None
        for j, label in enumerate(row):
            image_b = label_to_base_label_image(label)
            if label_to_description is not None:
                image_b = get_base_label_image_with_description(
                    base_label_image=image_b, label=label, description=label_to_description[label]
                )
            image_b = np.pad(
                image_b, pad_width=((60, 0), (0, 0), (0, 0)), constant_values=((0, 0), (0, 0), (0, 0)), mode="constant"
            )
            image_b = put_text_on_image(image=image_b, text=str(i + 1), fontsize=60, ymax=5, maximum_width=14)
            if total_image_row is None:
                total_image_row = image_b
            else:
                total_image_row = concat_images(
                    image_a=total_image_row,
                    image_b=image_b,
                    how="horizontally",
                    background_color_a=[0, 0, 0, 255] if j == 1 else None,
                    background_color_b=[0, 0, 0, 255],
                    thumbnail_size_a=(300, 300) if j == 1 else None,
                    thumbnail_size_b=(300, 300),
                )
            i += 1
        if total_image is None:
            total_image = total_image_row
        else:
            total_image = concat_images(image_a=total_image, image_b=total_image_row, how="vertically")
    return total_image


def get_image_binary_format(image: np.ndarray, format: str, **kwargs) -> str:
    image_io = io.BytesIO()
    image = np.array(image, dtype=np.uint8)
    Image.fromarray(image).save(image_io, format=format, **kwargs)
    return image_io.getvalue()


def get_image_b64(image: np.ndarray, format: str, **kwargs) -> str:
    image_format = get_image_binary_format(image, format, **kwargs)
    image_format_b64 = base64.b64encode(image_format).decode("utf-8")
    return image_format_b64


def draw_quadrangle_on_image(
    image: np.ndarray,
    points: Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int], Tuple[int, int]],
    color: Tuple[int, int, int] = None,
    thickness: int = 3,
    alpha: float = 0.3,
) -> np.ndarray:
    image = image.copy()
    rect = cv2.minAreaRect(np.array(points))
    box = cv2.boxPoints(rect)
    box = np.int0(box)
    cropped_image_zeros = np.ones_like(image)
    cv2.drawContours(image=cropped_image_zeros, contours=[box], contourIdx=0, color=color, thickness=thickness)
    colored_regions = cropped_image_zeros == color
    image[colored_regions] = (1 - alpha) * image[colored_regions] + alpha * cropped_image_zeros[colored_regions]
    return image


def get_thumbnail_resize(input_size: Tuple[int, int], target_size: Tuple[int, int]) -> Tuple[int, int]:
    # input_size = (width, height)
    x, y = map(math.floor, target_size)

    def round_aspect(number, key):
        return max(min(math.floor(number), math.ceil(number), key=key), 1)

    # preserve aspect ratio
    aspect = input_size[0] / input_size[1]
    if x / y >= aspect:
        x = round_aspect(y * aspect, key=lambda n: abs(aspect - n / y))
    else:
        y = round_aspect(x / aspect, key=lambda n: 0 if n == 0 else abs(aspect - x / n))
    size = (x, y)
    return size


def thumbnail_image(image: np.ndarray, size: Tuple[int, int], resample: Optional[int] = None) -> np.ndarray:
    image = Image.fromarray(image)
    new_width, new_height = get_thumbnail_resize((image.width, image.height), size)
    image = image.resize((new_width, new_height), resample=resample)
    image = np.array(image)
    return image


def tf_resize_with_pad(image: np.ndarray, target_width: int, target_height: int, constant_values: int) -> np.ndarray:
    import tensorflow as tf

    height_width_channels = tf.cast(tf.shape(image), tf.float32)
    height, width = height_width_channels[1], height_width_channels[0]
    ratio = tf.math.minimum(target_height / height, target_width / width)
    result_height = tf.cast(tf.math.round(ratio * height), tf.int32)
    result_width = tf.cast(tf.math.round(ratio * width), tf.int32)
    resized_tensor = tf.image.resize(image, [result_width, result_height])

    s = tf.shape(resized_tensor)
    paddings = [[(m - s[i]) / 2, (m - s[i]) / 2] for (i, m) in enumerate([target_height, target_width, 3])]
    return tf.pad(resized_tensor, paddings, "CONSTANT", constant_values=constant_values).numpy().astype(np.uint8)



import copy
import io
import json
from importlib.metadata import version
from pathlib import Path

import cv2
from packaging.version import Version

if Version(version("pydantic")) < Version("2.0.0"):
    from pydantic import BaseModel, Field, root_validator, validator
else:
    from pydantic.v1 import BaseModel, Field, validator, root_validator

from typing import Any, Dict, List, Optional, Tuple, Type, Union

import fsspec
import numpy as np
import PIL
from pathy import Pathy

from cv_pipeliner.utils.images import is_base64, open_image
from cv_pipeliner.utils.imagesize import get_image_size


def get_image_name(image_path) -> str:
    if isinstance(image_path, Pathy):
        return image_path.name
    elif isinstance(image_path, fsspec.core.OpenFile):
        return Pathy.fluid(image_path.path).name
    elif isinstance(image_path, str) or isinstance(image_path, bytes) or isinstance(image_path, io.BytesIO):
        return "bytes"
    elif isinstance(image_path, PIL.Image.Image):
        return "PIL.Image.Image"


def get_image_path_as_str(image_path) -> str:
    if isinstance(image_path, fsspec.core.OpenFile):
        protocol = image_path.fs.protocol
        if isinstance(protocol, tuple):
            protocol = protocol[0]
        prefix = f"{protocol}://"
        if protocol == "file":
            prefix = ""
        image_path_str = f"{prefix}{str(image_path.path)}"
    else:
        image_path_str = (
            str(image_path) if (image_path is not None and not isinstance(image_path, PIL.Image.Image)) else None
        )

    return image_path_str


def get_meta_image_size(
    image_path: Optional[Union[str, Path, fsspec.core.OpenFile, bytes, io.BytesIO, PIL.Image.Image]],
    image: Optional[np.ndarray],
    meta_height: Optional[int],
    meta_width: Optional[int],
    exif_transpose: bool = False,
):
    """
    Returns (width, height) of image without opening it fully.
    """
    if exif_transpose and image_path is not None:
        meta_width, meta_height = get_image_size(image_path, exif_transpose=exif_transpose)
    else:
        if meta_height is None or meta_width is None:
            if image is not None:
                meta_height, meta_width = image.shape[0:2]
            else:
                if image_path is None:
                    raise ValueError("(get_meta_image_size) Fields image_path or image are None")
                meta_width, meta_height = get_image_size(image_path, exif_transpose=exif_transpose)
        if image is not None:
            meta_height, meta_width = image.shape[0:2]
    return meta_width, meta_height


class BaseImageData(BaseModel):
    image_path: Optional[Union[str, Path, Pathy, fsspec.core.OpenFile, bytes, io.BytesIO, PIL.Image.Image]] = None
    image: Optional[np.ndarray] = Field(default=None, repr=False, exclude=True)
    label: Optional[str] = None
    keypoints: np.ndarray = Field(default_factory=lambda: np.array([]).astype(int).reshape((-1, 2)))
    mask: Union[
        Union[str, Path, Pathy, fsspec.core.OpenFile, bytes, io.BytesIO, PIL.Image.Image],  # path to mask image
        np.ndarray,  # mask image
        List[List[int]],  # [[x1, y1, x2, y2, ...], [x1, y1, x2, y2, ...], ...]
        List[List[Tuple[int, int]]],  # [[(x1, y1), (x2, y2), ...], [(x1, y1), (x2, y2), ...], ...]
        List[np.ndarray],  # [[x1, y1, x2, y2, ...], [x1, y1, x2, y2, ...], ...]
    ] = Field(default_factory=lambda: [], repr=False)
    detection_score: Optional[float] = Field(default=None)
    classification_score: Optional[float] = Field(default=None)
    top_n: Optional[int] = Field(default=None, repr=False)
    labels_top_n: Optional[Union[List[List[str]], np.ndarray, List[np.ndarray], List[str]]] = Field(
        default=None, repr=False
    )
    classification_scores_top_n: Optional[Union[List[List[float]], np.ndarray, List[np.ndarray], List[float]]] = Field(
        default=None, repr=False
    )
    additional_info: Dict[str, Any] = Field(default_factory=dict)
    meta_width: Optional[int] = None
    meta_height: Optional[int] = None

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True
        json_encoders = {
            Pathy: get_image_path_as_str,
            fsspec.core.OpenFile: get_image_path_as_str,
            np.ndarray: lambda x: x.tolist(),
        }
        smart_union = True
        copy_on_model_validation = "none"

    @validator("image_path", pre=True, allow_reuse=True)
    def parse_image_path(cls, image_path):
        if isinstance(image_path, Path) or (isinstance(image_path, str) and not is_base64(image_path)):
            image_path = Pathy.fluid(image_path)
        return image_path

    @validator("keypoints", pre=True, allow_reuse=True)
    def parse_keypoints(cls, keypoints):
        if keypoints is None:
            keypoints = []
        return np.array(keypoints).astype(int).reshape((-1, 2))

    @validator("mask", pre=True, allow_reuse=True)
    def parse_mask(cls, mask):
        if mask is None:
            return []
        if not isinstance(mask, list):
            if isinstance(mask, np.ndarray) and len(mask.shape) in [2, 3]:
                if len(mask.shape) == 2:
                    mask = mask[..., None]
                mask = (mask > 0).all(axis=2).astype(np.uint8)
                mask = cv2.copyMakeBorder(mask, 1, 1, 1, 1, cv2.BORDER_CONSTANT, value=0)
                polygons = cv2.findContours(mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE, offset=(-1, -1))
                polygons = polygons[0] if len(polygons) == 2 else polygons[1]
                polygons = [np.array(polygon, dtype=np.int32) for polygon in polygons]
                return polygons
        return [np.array(points, dtype=np.int32).reshape((-1, 2)) for points in mask if len(points) > 0]

    @validator("image", pre=True, allow_reuse=True)
    def parse_image(cls, image, values):
        if image is not None:
            image = np.array(image)
        return image

    @root_validator(pre=False, allow_reuse=True)
    def set_fields(cls, values: dict) -> dict:
        if values["image"] is not None:
            values["meta_width"], values["meta_height"] = get_meta_image_size(
                image_path=values["image_path"],
                image=values["image"],
                meta_height=values["meta_width"],
                meta_width=values["meta_width"],
            )
        return values

    def get_image_size(self, exif_transpose: bool = False) -> Tuple[int, int]:
        """
        Returns (width, height) of image without opening it fully.
        """
        meta_width, meta_height = get_meta_image_size(
            image_path=self.image_path,
            image=self.image,
            meta_height=self.meta_height,
            meta_width=self.meta_width,
            exif_transpose=exif_transpose,
        )
        if not exif_transpose:
            self.meta_width, self.meta_height = meta_width, meta_height
        return meta_width, meta_height

    @property
    def image_name(self):
        return get_image_name(self.image_path)

    @property
    def keypoints_n(self) -> List[Tuple[float]]:
        width, height = self.get_image_size()
        keypoints = self.keypoints.astype(float)
        keypoints[:, 0] /= width
        keypoints[:, 1] /= height
        return keypoints

    def open_image(
        self, inplace: bool = False, returns_none_if_empty: bool = False, exif_transpose: bool = False
    ) -> Optional[np.ndarray]:
        if self.image is not None and isinstance(self.image, np.ndarray):
            if not inplace:
                return self.image
            else:
                image = self.image.copy()
        elif isinstance(self.image, bytes) or isinstance(self.image, str):
            image = open_image(image=self.image, open_as_rgb=True, exif_transpose=exif_transpose)
        elif self.image_path is not None:
            image = open_image(image=self.image_path, open_as_rgb=True, exif_transpose=exif_transpose)
        else:
            if returns_none_if_empty:
                return None
            raise ValueError("Object doesn't have any image.")

        if inplace:
            self.image = image

        if image is not None:
            self.get_image_size(exif_transpose=exif_transpose)

        return image

    def json(self, include_image_path: bool = True, force_include_meta: bool = False, **kwargs) -> str:
        kwargs = kwargs.copy()
        exclude = kwargs.pop("exclude", set())
        if force_include_meta:
            self.get_image_size()  # write meta inplace if empty
        if not include_image_path:
            exclude.add("image_path")
        return super().json(exclude=exclude, exclude_none=kwargs.pop("exclude_none", True), **kwargs)

    @classmethod
    def from_json(
        cls,
        d: Optional[Union[Path, str, Dict[str, Any], fsspec.core.OpenFile]],
        image_path: Optional[Union[str, Path, fsspec.core.OpenFile, bytes, io.BytesIO, PIL.Image.Image]] = None,
    ):
        if cls is None:
            cls = BaseImageData

        if d is None:
            return cls(image_path=image_path)
        if isinstance(d, (str, Path)):
            try:
                d = json.loads(str(d))
            except Exception:
                with fsspec.open(d, "r") as f:
                    d = json.loads(f.read())
        elif isinstance(d, fsspec.core.OpenFile):
            with d as f:
                d = json.load(f)
        if image_path is not None:
            d["image_path"] = image_path

        return cls(**d)

    def is_empty(self):
        return self.image_path is None and self.image is None

    def open_mask(
        self,
        exif_transpose: bool = False,
    ) -> np.ndarray:
        if isinstance(self.mask, np.ndarray):
            return self.mask
        elif not isinstance(self.mask, list):
            mask = open_image(image=self.mask, open_as_rgb=True, exif_transpose=exif_transpose)
            return mask
        else:
            width, height = self.get_image_size(exif_transpose=exif_transpose)
            mask = np.zeros((height, width), dtype=np.uint8)
            mask = cv2.fillPoly(mask, self.mask, 255)
            return mask


class BboxData(BaseImageData):
    xmin: Optional[Union[int, float]] = None
    ymin: Optional[Union[int, float]] = None
    xmax: Optional[Union[int, float]] = None
    ymax: Optional[Union[int, float]] = None
    cropped_image: Optional[np.ndarray] = Field(default=None, repr=False)
    additional_bboxes_data: List["BboxData"] = Field(default_factory=list)

    @validator("xmin", "ymin", pre=True, allow_reuse=True)
    def parse_xmin(cls, v):
        if v is not None:
            v = max(0, v)
        return v

    @property
    def coords(self) -> Tuple[int, int, int, int]:
        return (round(self.xmin), round(self.ymin), round(self.xmax), round(self.ymax))

    @property
    def coords_n(self) -> Tuple[float, float, float, float]:
        width, height = self.get_image_size()
        return self.xmin / width, self.ymin / height, self.xmax / width, self.ymax / height

    @property
    def area(self) -> int:
        return (self.xmax - self.xmin + 1) * (self.ymax - self.ymin + 1)

    def coords_with_offset(
        self,
        xmin_offset: Union[int, float] = 0,
        ymin_offset: Union[int, float] = 0,
        xmax_offset: Union[int, float] = 0,
        ymax_offset: Union[int, float] = 0,
        source_image: np.ndarray = None,
    ) -> Tuple[int, int, int, int]:
        if source_image is not None:
            height, width = source_image.shape[0], source_image.shape[1]
        else:
            width, height = self.get_image_size()
        if isinstance(xmin_offset, float):
            assert 0 < xmin_offset and xmin_offset < 1
            xmin_offset = int((self.xmax - self.xmin) * xmin_offset)
        if isinstance(ymin_offset, float):
            assert 0 < ymin_offset and ymin_offset < 1
            ymin_offset = int((self.ymax - self.ymin) * ymin_offset)
        if isinstance(xmax_offset, float):
            assert 0 < xmax_offset and xmax_offset < 1
            xmax_offset = int((self.xmax - self.xmin) * xmax_offset)
        if isinstance(ymax_offset, float):
            assert 0 < ymax_offset and ymax_offset < 1
            ymax_offset = int((self.ymax - self.ymin) * ymax_offset)
        xmin_in_cropped_image = max(0, min(xmin_offset, self.xmin - xmin_offset))
        ymin_in_cropped_image = max(0, min(ymin_offset, self.ymin - ymin_offset))
        xmax_in_cropped_image = max(0, min(xmax_offset, width - self.xmax))
        ymax_in_cropped_image = max(0, min(ymax_offset, height - self.ymax))
        return (
            max(0, round(self.xmin - xmin_in_cropped_image)),
            max(0, round(self.ymin - ymin_in_cropped_image)),
            min(width - 1, round(self.xmax + xmax_in_cropped_image)),
            min(height - 1, round(self.ymax + ymax_in_cropped_image)),
        )

    def open_cropped_image(
        self,
        inplace: bool = False,
        source_image: np.ndarray = None,
        xmin_offset: Union[int, float] = 0,
        ymin_offset: Union[int, float] = 0,
        xmax_offset: Union[int, float] = 0,
        ymax_offset: Union[int, float] = 0,
        return_as_image_data: bool = False,
        image_data_cls: Type["ImageData"] = "ImageData",
    ) -> Optional[Union[np.ndarray, "BboxData"]]:
        if self.cropped_image is not None:
            if not inplace:
                return self.cropped_image
            else:
                cropped_image = self.cropped_image
        else:
            if source_image is not None:
                image = source_image
            else:
                image = self.open_image()
            self.meta_height, self.meta_width = image.shape[0:2]

            assert self.xmin < self.xmax and self.ymin < self.ymax

            xmin, ymin, xmax, ymax = self.coords_with_offset(
                xmin_offset, ymin_offset, xmax_offset, ymax_offset, source_image
            )
            cropped_image = image[ymin:ymax, xmin:xmax] if image is not None else None
        if inplace:
            self.cropped_image = cropped_image
        else:
            if return_as_image_data:
                keypoints = copy.deepcopy(self.keypoints)
                keypoints[:, 0] -= xmin
                keypoints[:, 1] -= ymin
                mask = copy.deepcopy(self.mask)
                for polygon in mask:
                    polygon[:, 0] -= xmin
                    polygon[:, 1] -= ymin

                copy_image = self.image  # Memory trick
                copy_cropped_image = self.cropped_image  # Memory trick
                self.image = None  # Memory trick
                self.cropped_image = None  # Memory trick
                additional_bboxes_data = copy.deepcopy(self.additional_bboxes_data)

                def crop_additional_bbox_data(bbox_data: BboxData):
                    bbox_data.image_path = None
                    bbox_data.image = cropped_image
                    bbox_data.xmin -= xmin
                    bbox_data.ymin -= ymin
                    bbox_data.xmax -= xmin
                    bbox_data.ymax -= ymin
                    bbox_data.keypoints[:, 0] -= xmin
                    bbox_data.keypoints[:, 1] -= ymin
                    if isinstance(bbox_data.mask, list):
                        for polygon in bbox_data.mask:
                            polygon[:, 0] -= xmin
                            polygon[:, 1] -= ymin
                    for additional_bbox_data in bbox_data.additional_bboxes_data:
                        crop_additional_bbox_data(additional_bbox_data)

                for additional_bbox_data in additional_bboxes_data:
                    crop_additional_bbox_data(additional_bbox_data)

                if image_data_cls == "ImageData":
                    from cv_pipeliner.core.data import ImageData

                    image_data_cls = ImageData

                image_data = image_data_cls(
                    image=cropped_image,
                    bboxes_data=additional_bboxes_data,
                    label=self.label,
                    keypoints=keypoints,
                    mask=mask,
                    additional_info=self.additional_info,
                )
                self.image = copy_image  # Memory trick
                self.cropped_image = copy_cropped_image  # Memory trick
                return image_data
            else:
                return cropped_image

    def count_children(self) -> int:
        global counting
        counting = 0

        def counts(bbox_data: BboxData):
            global counting
            counting += len(bbox_data.additional_bboxes_data)
            for additional_bbox_data in bbox_data.additional_bboxes_data:
                counts(additional_bbox_data)

        for additional_bbox_data in self.additional_bboxes_data:
            counts(additional_bbox_data)

        return counting

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        if name in ["image_path", "image", "meta_width", "meta_height"]:

            def change_images_in_bbox_data(bbox_data: BboxData):
                bbox_data.__setattr__(name, value)
                # for additional_bbox_data in bbox_data.additional_bboxes_data:
                #     change_images_in_bbox_data(additional_bbox_data)

            # if hasattr(self, "additional_bboxes_data"):
            #     for additional_bbox_data in self.additional_bboxes_data:
            #         change_images_in_bbox_data(additional_bbox_data)

    def open_cropped_mask(
        self,
        xmin_offset: Union[int, float] = 0,
        ymin_offset: Union[int, float] = 0,
        xmax_offset: Union[int, float] = 0,
        ymax_offset: Union[int, float] = 0,
        exif_transpose: bool = False,
        source_image: np.ndarray = None,
    ) -> np.ndarray:
        xmin, ymin, xmax, ymax = self.coords_with_offset(
            xmin_offset, ymin_offset, xmax_offset, ymax_offset, source_image
        )
        if self.mask is None or isinstance(self.mask, np.ndarray):
            return self.mask
        elif not isinstance(self.mask, list):
            mask = open_image(image=self.mask, open_as_rgb=True, exif_transpose=exif_transpose)
            mask_height, mask_width = mask.shape[0:2]
            if (mask_width, mask_height) == self.get_image_size(exif_transpos=exif_transpose):
                mask = mask[ymin:ymax, xmin:xmax]
            return mask
        else:
            width, height = (xmax - xmin), (ymax - ymin)
            polygons = copy.deepcopy(self.mask)
            for polygon in polygons:
                polygon[:, 0] = np.clip(polygon[:, 0] - xmin, 0, width - 1)
                polygon[:, 1] = np.clip(polygon[:, 1] - ymin, 0, height - 1)
            mask = np.zeros((height, width), dtype=np.uint8)
            mask = cv2.fillPoly(mask, polygons, 255)
            return mask


class ImageData(BaseImageData):
    bboxes_data: List[BboxData] = Field(default_factory=list)

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        if name in ["image_path", "image", "meta_width", "meta_height"]:
            if hasattr(self, "bboxes_data"):
                for bbox_data in self.bboxes_data:
                    bbox_data.__setattr__(name, value)

    def find_bbox_data_by_coords(self, xmin: int, ymin: int, xmax: int, ymax: int) -> BboxData:
        bboxes_data_coords = [bbox_data.coords for bbox_data in self.bboxes_data]
        return self.bboxes_data[bboxes_data_coords.index((xmin, ymin, xmax, ymax))]

    def count_children_bboxes_data(self) -> int:
        count = 0
        for bbox_data in self.bboxes_data:
            count += 1 + bbox_data.count_children()
        return count

    def open_mask(
        self,
        exif_transpose: bool = False,
        include_bboxes_data: bool = True,
        include_additional_bboxes_data: bool = True,
        additional_bboxes_data_depth: Optional[int] = None,
    ) -> np.ndarray:
        mask = BaseImageData.open_mask(self, exif_transpose=exif_transpose)
        if include_additional_bboxes_data:
            from cv_pipeliner.utils.images_datas import (
                flatten_additional_bboxes_data_in_image_data,
            )

            bboxes_data = flatten_additional_bboxes_data_in_image_data(
                self, additional_bboxes_data_depth=additional_bboxes_data_depth
            ).bboxes_data
        else:
            bboxes_data = self.bboxes_data
        if include_bboxes_data:
            for bbox_data in bboxes_data:
                new_mask = bbox_data.open_mask(exif_transpose=exif_transpose)
                mask[new_mask > 0] = new_mask[new_mask > 0]
        return mask



import json
import tempfile
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Type, Union

import fsspec
import numpy as np

from cv_pipeliner.core.inference_model import get_preprocess_input_from_script_file
from cv_pipeliner.inference_models.detection.core import (
    DetectionInput,
    DetectionModel,
    DetectionModelSpec,
    DetectionOutput,
)


class YOLOv8_ModelSpec(DetectionModelSpec):
    model_name: Optional[str] = None
    model_path: Optional[Union[str, Path]] = None  # noqa: F821
    class_names: Optional[Union[List[str], str, Path]] = None
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    device: str = None
    force_reload: bool = False

    @property
    def inference_model_cls(self) -> Type["YOLOv8_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.yolov8 import YOLOv8_DetectionModel

        return YOLOv8_DetectionModel


class YOLOv8_DetectionModel(DetectionModel):
    def __init__(self, model_spec: YOLOv8_ModelSpec):
        """YOLOv8 model initialization

        Args:
            model_spec (YOLOv8_ModelSpec): YOLOv8 Model specification

        Raises:
            ValueError: if passed wrong data type of model_spec
        """
        super().__init__(model_spec)

        # Loading classes names and save as attribute
        if model_spec.class_names is not None:
            if isinstance(model_spec.class_names, str) or isinstance(model_spec.class_names, Path):
                with fsspec.open(model_spec.class_names, "r", encoding="utf-8") as out:
                    self.class_names = np.array(json.load(out))
            else:
                self.class_names = np.array(model_spec.class_names)
        else:
            self.class_names = None

        # Loading preprocessing function
        if isinstance(model_spec.preprocess_input, str) or isinstance(model_spec.preprocess_input, Path):
            self._preprocess_input = get_preprocess_input_from_script_file(script_file=model_spec.preprocess_input)
        else:
            if model_spec.preprocess_input is None:
                # For numpy.ndarray inputs, YOLOv8 expects BGR format
                # https://github.com/ultralytics/ultralytics/issues/2575
                self._preprocess_input = lambda input: [image[:, :, ::-1] for image in input]
            else:
                self._preprocess_input = model_spec.preprocess_input

        # Loading model
        if isinstance(model_spec, YOLOv8_ModelSpec):
            self._load_yolov8_model(model_spec)
            self._raw_predict_images = self._raw_predict_images_torch
        else:
            raise ValueError(f"ObjectDetectionAPI_Model got unknown DetectionModelSpec: {type(model_spec)}")

    def _load_yolov8_model(self, model_spec: YOLOv8_ModelSpec):
        """YOLOv8 model initialization

        Args:
            model_spec (YOLOv8_ModelSpec): YOLOv8 Model specification

        Raises:
            ValueError: If model_name and model_path is not specified
        """
        if model_spec.model_name is None and model_spec.model_path is None:
            raise ValueError("Please, specify model name or weights path for loading model")

        if model_spec.model_path is not None:
            temp_file = tempfile.NamedTemporaryFile(suffix=".pt")
            with fsspec.open(model_spec.model_path, "rb") as src:
                temp_file.write(src.read())
            model_path_tmp = Path(temp_file.name)
            from ultralytics import YOLO

            self.model = YOLO(model_path_tmp)
        else:
            self.model = YOLO(model_spec.model_name)

        if model_spec.device is not None:
            self.model = self.model.to(model_spec.device)

    def _raw_predict_images_torch(
        self, input: DetectionInput, score_threshold: float
    ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray], List[str]]:
        """Private method to run pytorch model inference and return raw results

        Args:
            input (DetectionInput): list of images
            score_threshold (float): model confidence threshold

        Returns:
            Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray], List[str]]: _description_
        """

        predictions = self.model.predict(
            input,
            verbose=False,
            save_conf=True,
            conf=score_threshold,
            retina_masks=True,
        )
        raw_boxes, raw_keypoints, raw_masks, raw_scores, raw_labels = [], [], [], [], []
        for prediction in predictions:
            raw_boxes.append(prediction.boxes.xyxy.data.cpu().numpy())
            if prediction.masks is not None:
                all_keypoints = prediction.masks.xy
                raw_masks.append([[kp] for kp in all_keypoints])
            else:
                raw_masks.append([[[]] for _ in range(len(raw_boxes[-1])) for _ in range(len(raw_boxes))])
            raw_keypoints.append(np.array([]).reshape(len(raw_boxes[-1]), 0, 2))  # TODO: add keypoints support
            raw_labels.append(prediction.boxes.cls.data.cpu().numpy())
            raw_scores.append(prediction.boxes.conf.data.cpu().numpy())

        return raw_boxes, raw_keypoints, raw_masks, raw_scores, raw_labels

    def predict(
        self,
        input: DetectionInput,
        score_threshold: float,
        classification_top_n: int = 1,
    ) -> DetectionOutput:
        """Method to run model inference

        Args:
            input (DetectionInput): list of images
            score_threshold (float): model confidence threshold
            classification_top_n (int, optional): .... Defaults to None.

        Returns:
            DetectionOutput: List of boxes, keypoints, scores, classes
        """
        input = self._preprocess_input(input)
        # Try to rebatch to batches of one size due to https://github.com/ultralytics/ultralytics/issues/15430
        size_to_images = {}
        size_to_idxs = {}
        for idx, image in enumerate(input):
            size = tuple(image.shape[0:3])
            if size not in size_to_images:
                size_to_images[size] = []
                size_to_idxs[size] = []
            size_to_images[size].append(image)
            size_to_idxs[size].append(idx)

        idx_to_results = {}
        for size, images in size_to_images.items():
            (
                size_raw_bboxes,
                size_raw_keypoints,
                size_raw_masks,
                size_raw_scores,
                size_raw_classes,
            ) = self._raw_predict_images(images, score_threshold)
            for i, idx in enumerate(size_to_idxs[size]):
                idx_to_results[idx] = (
                    size_raw_bboxes[i],
                    size_raw_keypoints[i],
                    size_raw_masks[i],
                    size_raw_scores[i],
                    size_raw_classes[i],
                )
        results = [idx_to_results[idx] for idx in range(len(input))]
        raw_bboxes, raw_keypoints, raw_masks, raw_scores, raw_classes = zip(*results)

        if self.class_names is not None:
            if classification_top_n > 1:
                raise NotImplementedError("Not impelemented for classification_top_n > 1")
            class_names_top_n = [
                [
                    [class_name for i in range(classification_top_n)]
                    for class_name in self.class_names[classes.astype(np.int32)]
                ]
                for classes in raw_classes
            ]
            classes_scores_top_n = [[[score] for score in scores] for scores in raw_scores]
        else:
            class_names_top_n = [[None for _ in range(classification_top_n)] for _ in raw_classes]
            classes_scores_top_n = [[score for _ in range(classification_top_n)] for score in raw_scores]

        n_pred_bboxes = [image_boxes.tolist() for image_boxes in raw_bboxes]
        n_pred_keypoints = [np.array(k_keypoints).round().astype(np.int32).tolist() for k_keypoints in raw_keypoints]
        n_pred_masks = [
            [
                [np.array(polygon).round().astype(np.int32).tolist() for polygon in k_polygons]
                for k_polygons in n_k_polygons
            ]
            for n_k_polygons in raw_masks
        ]
        n_pred_scores = [image_scores.tolist() for image_scores in raw_scores]
        n_pred_class_names_top_k = class_names_top_n
        n_pred_scores_top_k = classes_scores_top_n
        return (
            n_pred_bboxes,
            n_pred_keypoints,
            n_pred_masks,
            n_pred_scores,
            n_pred_class_names_top_k,
            n_pred_scores_top_k,
        )

    def preprocess_input(self, input: DetectionInput) -> DetectionInput:
        return self._preprocess_input(input)

    @property
    def input_size(self) -> int:
        return self.model.imgsz



__version__ = "0.20.2"

from cv_pipeliner.batch_generators.bbox_data import BatchGeneratorBboxData
from cv_pipeliner.batch_generators.image_data import BatchGeneratorImageData
from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.data_converters.brickit import BrickitDataConverter
from cv_pipeliner.data_converters.json import JSONDataConverter
from cv_pipeliner.data_converters.supervisely import SuperviselyDataConverter
from cv_pipeliner.data_converters.yolo import YOLODataConverter, YOLOMasksDataConverter
from cv_pipeliner.inference_models.classification.tensorflow import (
    TensorFlow_ClassificationModelSpec,
    TensorFlow_ClassificationModelSpec_TFServing,
)
from cv_pipeliner.inference_models.detection.detectron2 import Detectron2_ModelSpec
from cv_pipeliner.inference_models.detection.object_detection_api import (
    ObjectDetectionAPI_KFServing,
    ObjectDetectionAPI_ModelSpec,
    ObjectDetectionAPI_pb_ModelSpec,
    ObjectDetectionAPI_TFLite_ModelSpec,
)
from cv_pipeliner.inference_models.detection.yolov5 import (
    YOLOv5_ModelSpec,
    YOLOv5_TFLite_ModelSpec,
    YOLOv5_TFLiteWithNMS_ModelSpec,
)
from cv_pipeliner.inference_models.detection.yolov8 import YOLOv8_ModelSpec
from cv_pipeliner.inference_models.embedder.pytorch import PyTorch_EmbedderModelSpec
from cv_pipeliner.inference_models.embedder.tensorflow import (
    TensorFlow_EmbedderModelSpec,
)
from cv_pipeliner.inference_models.keypoints_regressor.mmpose import (
    MMPose_KeypointsRegressorModelSpec_TFLite,
)
from cv_pipeliner.inference_models.keypoints_regressor.tensorflow import (
    TensorFlow_KeypointsRegressorModelSpec,
    TensorFlow_KeypointsRegressorModelSpec_TFServing,
)
from cv_pipeliner.inference_models.pipeline import PipelineModel, PipelineModelSpec
from cv_pipeliner.inferencers.classification import ClassificationInferencer
from cv_pipeliner.inferencers.detection import DetectionInferencer
from cv_pipeliner.inferencers.keypoints_regressor import KeypointsRegressorInferencer
from cv_pipeliner.inferencers.pipeline import PipelineInferencer
from cv_pipeliner.metrics.classification import get_df_classification_metrics
from cv_pipeliner.metrics.detection import get_df_detection_metrics
from cv_pipeliner.metrics.image_data_matching import (
    BboxDataMatching,
    ImageDataMatching,
    intersection_over_union,
    pairwise_intersection_over_union,
)
from cv_pipeliner.metrics.pipeline import get_df_pipeline_metrics
from cv_pipeliner.utils.fiftyone import FifyOneSession
from cv_pipeliner.utils.images import (
    concat_images,
    draw_rectangle,
    put_text_on_image,
    thumbnail_image,
)
from cv_pipeliner.utils.images_datas import (
    apply_perspective_transform_to_image_data,
    combine_mask_polygons_to_one_polygon,
    crop_image_data,
    flatten_additional_bboxes_data_in_image_data,
    get_perspective_matrix_for_base_keypoints,
    non_max_suppression_image_data,
    non_max_suppression_image_data_using_tf,
    resize_image_data,
    rotate_image_data,
    split_image_by_grid,
    thumbnail_image_data,
    uncrop_bboxes_data,
)
from cv_pipeliner.utils.imagesize import get_image_size
from cv_pipeliner.visualizers.core.image_data import visualize_image_data
from cv_pipeliner.visualizers.core.image_data_matching import (
    visualize_image_data_matching_side_by_side,
)



import abc
from typing import List, Tuple, Type

import numpy as np

from cv_pipeliner.core.inference_model import InferenceModel, ModelSpec

Bbox = Tuple[int, int, int, int]  # (xmin, ymin, xmax, ymax)
Score = float
Class = str

Bboxes = List[Bbox]
DetectionScores = List[Score]
ClassificationScores = List[Score]
Classes = List[Class]
Keypoints = List[List[Tuple[int, int]]]
Mask = List[List[List[Tuple[int, int]]]]

DetectionInput = List[np.ndarray]
DetectionOutput = Tuple[
    List[Bboxes],
    List[Keypoints],
    List[Mask],
    List[DetectionScores],
    List[Classes],
    List[ClassificationScores],
]


class DetectionModelSpec(ModelSpec):
    class_names: List[str] = None  # optional

    @abc.abstractproperty
    def inference_model_cls(self) -> Type["DetectionModel"]:
        pass

    def load_detection_inferencer(self) -> "DetectionInferencer":
        from cv_pipeliner.inferencers.detection import DetectionInferencer

        return DetectionInferencer(self.load())


class DetectionModel(InferenceModel):
    @abc.abstractmethod
    def __init__(self, model_spec: DetectionModelSpec):
        assert isinstance(model_spec, DetectionModelSpec)
        super().__init__(model_spec)

    @abc.abstractmethod
    def predict(
        self, input: DetectionInput, score_threshold: float, classification_top_n: int = None
    ) -> DetectionOutput:
        pass

    @abc.abstractmethod
    def preprocess_input(self, input):
        pass

    @abc.abstractproperty
    def input_size(self) -> int:
        pass



import tempfile
from pathlib import Path

import fsspec
from pathy import Pathy

from cv_pipeliner.logging import logger


def copy_files_from_directory_to_temp_directory(directory: str) -> tempfile.TemporaryDirectory:
    directory_openfile = fsspec.open(directory)
    directory = Pathy.fluid(directory)
    temp_dir = tempfile.TemporaryDirectory()
    temp_dir_path = Path(temp_dir.name)

    for some_file in fsspec.open_files(str(directory / "**"), "rb"):
        some_file_path = Pathy.fluid(some_file.path)
        relative_filepath = some_file_path.relative_to(directory_openfile.path)
        filepath = temp_dir_path / relative_filepath
        (filepath.parent).mkdir(exist_ok=True, parents=True)
        logger.info(f"copy {some_file_path} => {filepath}")
        with open(filepath, "wb") as out:
            with some_file as src:
                out.write(src.read())

    return temp_dir



from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from sklearn.metrics import multilabel_confusion_matrix

from cv_pipeliner.core.data import BboxData


def _add_metrics_to_dict(
    classification_metrics: Dict,
    labels: List[str],
    tops_n: List[int],
    prefix_caption: str = "",
    postfix_caption: str = "",
    count_mean_expected_steps: bool = False,
):
    supports = [classification_metrics[class_name]["support"] for class_name in labels]
    support = np.sum(supports)
    TP = np.sum([classification_metrics[class_name]["TP"] for class_name in labels])
    FP = np.sum([classification_metrics[class_name]["FP"] for class_name in labels])
    FN = np.sum([classification_metrics[class_name]["FN"] for class_name in labels])
    accuracy = TP / max(TP + FP + FN, 1e-6)
    micro_average_precision = TP / max(TP + FP, 1e-6)
    micro_average_recall = TP / max(TP + FN, 1e-6)
    micro_average_f1_score = (
        2 * micro_average_precision * micro_average_recall / (max(micro_average_precision + micro_average_recall, 1e-6))
    )
    precisions = [classification_metrics[class_name]["precision"] for class_name in labels]
    recalls = [classification_metrics[class_name]["recall"] for class_name in labels]
    f1_scores = [classification_metrics[class_name]["f1_score"] for class_name in labels]
    macro_average_precision = np.average(precisions) if len(precisions) > 0 else np.nan
    weighted_average_precision = np.average(precisions, weights=supports) if len(precisions) > 0 else np.nan
    macro_average_recall = np.average(recalls) if len(recalls) > 0 else np.nan
    weighted_average_recall = np.average(recalls, weights=supports) if len(recalls) > 0 else np.nan
    macro_average_f1_score = np.average(f1_scores) if len(f1_scores) > 0 else np.nan
    weighted_average_f1_score = np.average(f1_scores, weights=supports) if len(f1_scores) > 0 else np.nan
    sum_support = np.sum(supports)
    classification_metrics[f"{prefix_caption}accuracy{postfix_caption}"] = {
        "support": support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "value": accuracy,
    }
    classification_metrics[f"{prefix_caption}micro_average{postfix_caption}"] = {
        "support": sum_support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "precision": micro_average_precision,
        "recall": micro_average_recall,
        "f1_score": micro_average_f1_score,
    }
    classification_metrics[f"{prefix_caption}macro_average{postfix_caption}"] = {
        "support": sum_support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "precision": macro_average_precision,
        "recall": macro_average_recall,
        "f1_score": macro_average_f1_score,
    }
    classification_metrics[f"{prefix_caption}weighted_average{postfix_caption}"] = {
        "support": sum_support,
        "TP": TP,
        "FP": FP,
        "FN": FN,
        "precision": weighted_average_precision,
        "recall": weighted_average_recall,
        "f1_score": weighted_average_f1_score,
    }
    if count_mean_expected_steps:
        not_nan_labels = [
            class_name
            for class_name in labels
            if not np.isnan(classification_metrics[class_name]["mean_expected_steps"])
        ]
        mean_expected_steps = [
            classification_metrics[class_name]["mean_expected_steps"] for class_name in not_nan_labels
        ]
        mean_expected_steps_supports = [classification_metrics[class_name]["support"] for class_name in not_nan_labels]
        macro_average_mean_expected_steps = np.average(mean_expected_steps) if len(mean_expected_steps) > 0 else np.nan
        weighted_average_mean_expected_steps = (
            np.average(mean_expected_steps, weights=mean_expected_steps_supports)
            if len(mean_expected_steps) > 0
            else np.nan
        )
        classification_metrics[f"{prefix_caption}macro_average{postfix_caption}"][
            "mean_expected_steps"
        ] = macro_average_mean_expected_steps
        classification_metrics[f"{prefix_caption}weighted_average{postfix_caption}"][
            "mean_expected_steps"
        ] = weighted_average_mean_expected_steps

    for top_n in tops_n:
        if top_n == 1:
            continue
        precisions_top_n = [classification_metrics[class_name][f"precision@{top_n}"] for class_name in labels]
        recalls_top_n = [classification_metrics[class_name][f"recall@{top_n}"] for class_name in labels]
        f1_score_top_n = [classification_metrics[class_name][f"f1_score@{top_n}"] for class_name in labels]
        macro_average_precision_top_n = np.average(precisions_top_n) if len(precisions_top_n) > 0 else np.nan
        weighted_average_precision_top_n = (
            np.average(precisions_top_n, weights=supports) if len(precisions_top_n) > 0 else np.nan
        )
        macro_average_recall_top_n = np.average(recalls_top_n) if len(recalls_top_n) > 0 else np.nan
        weighted_average_recall_top_n = (
            np.average(recalls_top_n, weights=supports) if len(recalls_top_n) > 0 else np.nan
        )
        macro_average_f1_score_top_n = np.average(f1_score_top_n) if len(f1_score_top_n) > 0 else np.nan
        weighted_average_f1_score_top_n = (
            np.average(f1_score_top_n, weights=supports) if len(f1_score_top_n) > 0 else np.nan
        )
        classification_metrics[f"{prefix_caption}macro_average{postfix_caption}"][
            f"precision@{top_n}"
        ] = macro_average_precision_top_n
        classification_metrics[f"{prefix_caption}weighted_average{postfix_caption}"][
            f"precision@{top_n}"
        ] = weighted_average_precision_top_n
        classification_metrics[f"{prefix_caption}macro_average{postfix_caption}"][
            f"recall@{top_n}"
        ] = macro_average_recall_top_n
        classification_metrics[f"{prefix_caption}weighted_average{postfix_caption}"][
            f"recall@{top_n}"
        ] = weighted_average_recall_top_n
        classification_metrics[f"{prefix_caption}macro_average{postfix_caption}"][
            f"f1_score@{top_n}"
        ] = macro_average_f1_score_top_n
        classification_metrics[f"{prefix_caption}weighted_average{postfix_caption}"][
            f"f1_score@{top_n}"
        ] = weighted_average_f1_score_top_n


def get_precision_and_recall_top_n(
    true_labels: List[str], pred_labels_top_n: List[List[str]], label: str, top_n: int
) -> Tuple[int, int]:
    # Precision@k = (# of recommended items @k that are relevant) / (# of recommended items @k)
    # Recall@k = (# of recommended items @k that are relevant) / (total # of relevant items)
    recommended_relevant_items = 0
    recommended_items = 0
    relevant_items = 0
    for true_label, pred_label_top_n in zip(true_labels, pred_labels_top_n):
        if label in pred_label_top_n[0:top_n]:
            recommended_items += 1
        if true_label == label:
            relevant_items += 1
            if label in pred_label_top_n[0:top_n]:
                recommended_relevant_items += 1

    precision_top_n = recommended_relevant_items / max(recommended_items, 1e-6)
    recall_top_n = recommended_relevant_items / max(relevant_items, 1e-6)

    return precision_top_n, recall_top_n


def get_mean_expected_steps(
    n_true_labels: List[List[str]],
    n_pred_labels_top_n: List[List[List[str]]],
    n_pred_scores_top_n: List[List[List[float]]],
    label: str,
    step_penalty: int,
) -> Tuple[int, int]:
    n_steps = []
    for true_labels, pred_labels_top_n, pred_scores_top_n in zip(
        n_true_labels, n_pred_labels_top_n, n_pred_scores_top_n
    ):
        idxs_by_label = pred_labels_top_n == label
        pred_scores_top_n_by_label = pred_scores_top_n[idxs_by_label]
        idxs_sorted = np.argsort((-1) * pred_scores_top_n_by_label)
        true_labels_sorted = true_labels[idxs_sorted]
        steps = np.where(label == true_labels_sorted)[0]
        nth_step_penalty = step_penalty if len(np.where(label == true_labels)[0]) > 0 else None
        steps = min(step_penalty, np.min(steps) + 1) if len(steps) > 0 else nth_step_penalty
        if steps is not None:
            n_steps.append(steps)
    mean_expected_steps = np.mean(n_steps) if len(n_steps) > 0 else 0
    return mean_expected_steps


def get_df_classification_metrics(
    n_true_bboxes_data: List[List[BboxData]],
    n_pred_bboxes_data: List[List[BboxData]],
    pseudo_class_names: List[str],
    known_class_names: List[str] = None,
    tops_n: List[int] = [1],
    step_penalty: int = 20,
) -> pd.DataFrame:
    # We use pipeline metrics for it:

    assert len(n_true_bboxes_data) == len(n_true_bboxes_data)
    true_bboxes_data = np.array([bbox_data for bboxes_data in n_true_bboxes_data for bbox_data in bboxes_data])
    pred_bboxes_data = np.array([bbox_data for bboxes_data in n_pred_bboxes_data for bbox_data in bboxes_data])
    assert len(true_bboxes_data) == len(pred_bboxes_data)
    true_labels = np.array([bbox_data.label for bbox_data in true_bboxes_data])
    pred_labels = np.array([bbox_data.label for bbox_data in pred_bboxes_data])
    pred_labels_top_n = np.array(
        [
            bbox_data.labels_top_n if bbox_data.labels_top_n is not None else [bbox_data.label]
            for bbox_data in pred_bboxes_data
        ]
    )
    min_tops_n_from_pred_bboxes_data = min(
        [bbox_data.top_n if bbox_data.top_n is not None else 1 for bbox_data in pred_bboxes_data]
    )
    assert max(tops_n) <= min_tops_n_from_pred_bboxes_data
    n_true_labels = [
        np.array([true_bbox_data.label for true_bbox_data in true_bboxes_data])
        for true_bboxes_data in n_true_bboxes_data
    ]
    n_pred_labels_top_n = [
        np.array(
            [
                pred_bbox_data.labels_top_n if pred_bbox_data.labels_top_n is not None else [pred_bbox_data.label]
                for pred_bbox_data in pred_bboxes_data
            ]
        )
        for pred_bboxes_data in n_pred_bboxes_data
    ]
    n_pred_scores_top_n = [
        np.array(
            [
                (
                    pred_bbox_data.classification_scores_top_n
                    if pred_bbox_data.classification_scores_top_n is not None
                    else [1.0]
                )
                for pred_bbox_data in pred_bboxes_data
            ]
        )
        for pred_bboxes_data in n_pred_bboxes_data
    ]

    all_class_names = np.unique(np.concatenate([true_labels, pred_labels])).tolist()
    class_names_without_pseudo_classes = list(set(all_class_names) - set(pseudo_class_names))
    MMM = multilabel_confusion_matrix(y_true=true_labels, y_pred=pred_labels, labels=all_class_names)
    classification_metrics = {}
    for idx, class_name in enumerate(all_class_names):
        support_by_class_name = np.sum(true_labels == class_name)
        TP_by_class_name = MMM[idx, 1, 1]
        FP_by_class_name = MMM[idx, 0, 1]
        FN_by_class_name = MMM[idx, 1, 0]
        precision_by_class_name = TP_by_class_name / max(TP_by_class_name + FP_by_class_name, 1e-6)
        recall_by_class_name = TP_by_class_name / max(TP_by_class_name + FN_by_class_name, 1e-6)
        f1_score_by_class_name = (
            2
            * precision_by_class_name
            * recall_by_class_name
            / (max(precision_by_class_name + recall_by_class_name, 1e-6))
        )
        classification_metrics[class_name] = {
            "support": support_by_class_name,
            "TP": TP_by_class_name,
            "FP": FP_by_class_name,
            "FN": FN_by_class_name,
            "precision": precision_by_class_name,
            "recall": recall_by_class_name,
            "f1_score": f1_score_by_class_name,
        }
        if max(tops_n) > 1:
            for top_n in tops_n:
                if top_n == 1:
                    continue
                precision_top_n, recall_top_n = get_precision_and_recall_top_n(
                    true_labels=true_labels, pred_labels_top_n=pred_labels_top_n, label=class_name, top_n=top_n
                )
                classification_metrics[class_name][f"precision@{top_n}"] = precision_top_n
                classification_metrics[class_name][f"recall@{top_n}"] = recall_top_n
                classification_metrics[class_name][f"f1_score@{top_n}"] = (2 * precision_top_n * recall_top_n) / max(
                    precision_top_n + recall_top_n, 1e-6
                )

    if known_class_names is not None:
        len_known_class_names = len(known_class_names)
        count_mean_expected_steps = min_tops_n_from_pred_bboxes_data >= len_known_class_names
        known_class_names = list(set(all_class_names).intersection(set(known_class_names)))
        known_class_names_without_pseudo_classes = list(set(known_class_names) - set(pseudo_class_names))
        if count_mean_expected_steps:
            for class_name in all_class_names:
                classification_metrics[class_name]["mean_expected_steps"] = get_mean_expected_steps(
                    n_true_labels=n_true_labels,
                    n_pred_labels_top_n=n_pred_labels_top_n,
                    n_pred_scores_top_n=n_pred_scores_top_n,
                    label=class_name,
                    step_penalty=step_penalty,
                )
        _add_metrics_to_dict(
            classification_metrics=classification_metrics,
            labels=known_class_names,
            tops_n=tops_n,
            prefix_caption="known_",
            count_mean_expected_steps=count_mean_expected_steps,
        )
        _add_metrics_to_dict(
            classification_metrics=classification_metrics,
            labels=known_class_names_without_pseudo_classes,
            tops_n=tops_n,
            prefix_caption="known_",
            postfix_caption="_without_pseudo_classes",
            count_mean_expected_steps=count_mean_expected_steps,
        )
    else:
        count_mean_expected_steps = False
    _add_metrics_to_dict(
        classification_metrics=classification_metrics,
        labels=all_class_names,
        tops_n=tops_n,
        prefix_caption="all_",
        count_mean_expected_steps=count_mean_expected_steps,
    )
    _add_metrics_to_dict(
        classification_metrics=classification_metrics,
        labels=class_names_without_pseudo_classes,
        tops_n=tops_n,
        prefix_caption="all_",
        postfix_caption="_without_pseudo_classes",
        count_mean_expected_steps=count_mean_expected_steps,
    )

    df_classification_metrics = pd.DataFrame(classification_metrics, dtype=object).T
    df_classification_metrics.sort_values(by="support", ascending=False, inplace=True)
    df_classification_metrics_MES_column = (
        ["mean_expected_steps"] if "mean_expected_steps" in df_classification_metrics.columns else []
    )
    df_classification_metrics_columns = (
        ["support"]
        + df_classification_metrics_MES_column
        + ["precision", "recall", "f1_score", "value"]
        + [
            item
            for sublist in [[f"precision@{top_n}", f"recall@{top_n}"] for top_n in tops_n if top_n > 1]
            for item in sublist
        ]
        + ["TP", "FP", "FN"]
    )
    df_classification_metrics = df_classification_metrics[df_classification_metrics_columns]

    if known_class_names is not None:
        df_classification_metrics.loc[all_class_names, "known"] = df_classification_metrics.loc[
            all_class_names
        ].index.isin(known_class_names)
        df_classification_metrics.loc[all_class_names, "pseudo"] = df_classification_metrics.loc[
            all_class_names
        ].index.isin(pseudo_class_names)

    return df_classification_metrics



import json
from pathlib import Path
from typing import Dict, List, Union

import fsspec
from pathy import Pathy

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.core.data_converter import DataConverter


class BrickitDataConverter(DataConverter):
    def get_annot_from_image_data(self, image_data: ImageData) -> Dict:
        annot = {"filename": image_data.image_name, "objects": []}
        image_data = self.filter_image_data(image_data)
        for bbox_data in image_data.bboxes_data:
            bbox_data_json = json.loads(bbox_data.json())
            obj = {
                "bbox": [
                    bbox_data_json["xmin"],
                    bbox_data_json["ymin"],
                    bbox_data_json["xmax"],
                    bbox_data_json["ymax"],
                ],
                "label": bbox_data_json["label"],
                **(
                    {key: bbox_data_json["additional_info"][key] for key in bbox_data_json["additional_info"]}
                    if "additional_info" in bbox_data_json
                    else {}
                ),
            }
            if bbox_data.labels_top_n is not None:
                obj["labels_top_n"] = bbox_data_json["labels_top_n"]

            annot["objects"].append(obj)
        if len(image_data.additional_info) > 0:
            for key in image_data.additional_info:
                annot[key] = image_data.additional_info[key]
        return annot

    @DataConverter.assert_images_data
    def get_images_data_from_annots(
        self, images_dir: Union[str, Path, fsspec.core.OpenFile], annots: Union[Path, str, Dict, fsspec.core.OpenFile]
    ) -> List[ImageData]:
        if isinstance(annots, str) or isinstance(annots, Path):
            with fsspec.open(annots, "r", encoding="utf8") as f:
                annots = json.load(f)
        elif isinstance(annots, fsspec.core.OpenFile):
            with annots as f:
                annots = json.load(f)
        images_dir = Pathy.fluid(images_dir)

        images_data = []
        for annot in annots:
            image_name = annot["filename"]
            additional_info = {}
            for key in annot:
                if key not in ["objects", "filename"]:
                    additional_info[key] = annot[key]
            bboxes_data = []
            for obj in annot["objects"]:
                if "bbox" in obj:
                    xmin, ymin, xmax, ymax = obj["bbox"]
                else:
                    xmin, ymin, xmax, ymax = obj
                xmin, ymin, xmax, ymax = int(xmin), int(ymin), int(xmax), int(ymax)
                label = obj["label"] if "label" in obj else None
                labels_top_n = obj["labels_top_n"] if "labels_top_n" in obj else None
                top_n = len(labels_top_n) if labels_top_n is not None else None
                keypoints = obj["keypoints"] if "keypoints" in obj else []
                bbox_additional_info = {}
                if isinstance(obj, dict):
                    for key in obj:
                        if key not in ["bbox", "label", "labels_top_n", "top_n"]:
                            bbox_additional_info[key] = obj[key]
                image_path = (images_dir / image_name) if image_name is not None else None
                bboxes_data.append(
                    BboxData(
                        image_path=image_path,
                        xmin=xmin,
                        ymin=ymin,
                        xmax=xmax,
                        ymax=ymax,
                        keypoints=keypoints,
                        label=label,
                        labels_top_n=labels_top_n,
                        top_n=top_n,
                        additional_info=bbox_additional_info,
                    )
                )

            images_data.append(
                ImageData(image_path=image_path, bboxes_data=bboxes_data, additional_info=additional_info)
            )

        return images_data

    @DataConverter.assert_image_data
    def get_image_data_from_annot(
        self, image_path: Union[str, Path], annot: Union[Path, str, Dict, fsspec.core.OpenFile]
    ) -> ImageData:
        if isinstance(annot, str) or isinstance(annot, Path):
            with fsspec.open(annot, "r", encoding="utf8") as f:
                annots = json.load(f)
        elif isinstance(annot, fsspec.core.OpenFile):
            with annot as f:
                annots = json.load(f)
        elif isinstance(annot, dict):
            annots = [annot]
        else:
            annots = annot
        image_path = Pathy.fluid(image_path)
        annots = [annot for annot in annots if annot["filename"] == image_path.name]

        return self.get_images_data_from_annots(images_dir=image_path.parent, annots=annots)[0]



import logging
import tarfile
from pathlib import Path
from typing import Union

import requests
from tqdm import tqdm

logger = logging.getLogger(__name__)


def download_and_extract_tar_gz_to_directory(url: str, directory: Union[str, Path]):
    directory = Path()
    directory.mkdir(parents=True, exist_ok=True)
    filepath = directory / "tempfile.tar.gz"

    logger.info(f"Downloading tar.gz archive from {url} ...")
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get("content-length", 0))
    with open(filepath, "wb") as file, tqdm(
        desc=str(filepath),
        total=total,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    ) as tbar:
        for data in resp.iter_content(chunk_size=1024):
            size = file.write(data)
            tbar.update(size)

    logger.info(f"Extracting all to {directory}...")
    tar = tarfile.open(filepath)
    tar.extractall(path=directory)
    tar.close()

    filepath.unlink()
    logger.info(f"Saved to {directory}")



from typing import Callable, List, Tuple, Union

import numpy as np
from tqdm import tqdm

from cv_pipeliner.batch_generators.bbox_data import BatchGeneratorBboxData
from cv_pipeliner.batch_generators.image_data import BatchGeneratorImageData
from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.core.inferencer import Inferencer
from cv_pipeliner.inference_models.keypoints_regressor.core import (
    KeypointsRegressorModel,
)


class KeypointsRegressorInferencer(Inferencer):
    def __init__(self, model: KeypointsRegressorModel):
        assert isinstance(model, KeypointsRegressorModel)
        super().__init__(model)

    def _postprocess_images_data(
        self,
        images_data: List[ImageData],
        pred_n_keypoints: List[List[Tuple[float, float]]],
        open_images_in_images_data: bool,
    ) -> List[ImageData]:
        images_data_res = []
        pred_n_keypoints = np.array(pred_n_keypoints)
        for image_data, pred_keypoints in zip(images_data, pred_n_keypoints):
            image = image_data.image if open_images_in_images_data else None
            images_data_res.append(
                ImageData(
                    image_path=image_data.image_path,
                    image=image,
                    bboxes_data=image_data.bboxes_data,
                    label=image_data.label,
                    keypoints=pred_keypoints,
                    additional_info=image_data.additional_info,
                    meta_width=image_data.meta_width,
                    meta_height=image_data.meta_height,
                )
            )

        return images_data_res

    def _predict_images_data(
        self,
        images_data_gen: BatchGeneratorImageData,
        open_images_in_images_data: bool = False,
        disable_tqdm: bool = False,
        progress_callback: Callable[[int], None] = lambda progress: None,
    ):
        pred_images_data = []
        with tqdm(total=len(images_data_gen.data), disable=disable_tqdm) as pbar:
            for images_data in images_data_gen:
                input = [image_data.image for image_data in images_data]
                pred_n_keypoints = self.model.predict(input=input)
                pred_images_data.extend(
                    self._postprocess_images_data(
                        images_data=images_data,
                        pred_n_keypoints=pred_n_keypoints,
                        open_images_in_images_data=open_images_in_images_data,
                    )
                )
                pbar.update(len(images_data))
                progress_callback(pbar.n)

        return pred_images_data

    def _postprocess_predictions_bboxes_data(
        self,
        bboxes_data: List[BboxData],
        pred_n_keypoints: List[List[Tuple[float, float]]],
        open_cropped_images_in_bboxes_data: bool,
    ) -> List[List[BboxData]]:
        bboxes_data_res = []
        pred_n_keypoints = np.array(pred_n_keypoints)
        for bbox_data, pred_keypoints in zip(bboxes_data, pred_n_keypoints):
            cropped_image = bbox_data.cropped_image if open_cropped_images_in_bboxes_data else None
            bboxes_data_res.append(
                BboxData(
                    image_path=bbox_data.image_path,
                    cropped_image=cropped_image,
                    xmin=bbox_data.xmin,
                    ymin=bbox_data.ymin,
                    xmax=bbox_data.xmax,
                    ymax=bbox_data.ymax,
                    detection_score=bbox_data.detection_score,
                    label=bbox_data.label,
                    keypoints=pred_keypoints,
                    classification_score=bbox_data.classification_score,
                    top_n=bbox_data.top_n,
                    labels_top_n=bbox_data.labels_top_n,
                    classification_scores_top_n=bbox_data.classification_scores_top_n,
                )
            )

        return bboxes_data_res

    def _split_chunks(self, _list: List, shapes: List[int]) -> List:
        cnt = 0
        chunks = []
        for shape in shapes:
            chunks.append(_list[cnt : cnt + shape])
            cnt += shape
        return chunks

    def _predict_bboxes_data(
        self,
        n_bboxes_data_gen: BatchGeneratorBboxData,
        open_images_in_bboxes_data: bool = False,
        disable_tqdm: bool = False,
        progress_callback: Callable[[int], None] = lambda progress: None,
    ):
        pred_bboxes_data = []
        with tqdm(total=len(n_bboxes_data_gen.data), disable=disable_tqdm) as pbar:
            for bboxes_data in n_bboxes_data_gen:
                input = [bbox_data.cropped_image for bbox_data in bboxes_data]
                pred_n_keypoints = self.model.predict(input=input)
                pred_bboxes_data.extend(
                    self._postprocess_predictions_bboxes_data(
                        bboxes_data=bboxes_data,
                        pred_n_keypoints=pred_n_keypoints,
                        open_cropped_images_in_bboxes_data=open_images_in_bboxes_data,
                    )
                )
                pbar.update(len(bboxes_data))
                if progress_callback is not None:
                    progress_callback(pbar.n)

        n_pred_bboxes_data = self._split_chunks(pred_bboxes_data, n_bboxes_data_gen.shapes)
        return n_pred_bboxes_data

    def predict(
        self,
        data_gen: Union[BatchGeneratorImageData, BatchGeneratorBboxData],
        open_images_in_data: bool = False,
        disable_tqdm: bool = False,
        progress_callback: Callable[[int], None] = lambda progress: None,
        batch_size_default: int = 32,
    ) -> Union[List[ImageData], List[List[BboxData]]]:
        if isinstance(data_gen, list):
            if all(isinstance(d, ImageData) for d in data_gen):
                data_gen = BatchGeneratorImageData(data_gen, batch_size=batch_size_default)
            elif all(isinstance(d, BboxData) for d in data_gen):
                data_gen = BatchGeneratorBboxData(data_gen, batch_size=batch_size_default)
        if isinstance(data_gen, BatchGeneratorImageData):
            return self._predict_images_data(
                images_data_gen=data_gen,
                open_images_in_images_data=open_images_in_data,
                disable_tqdm=disable_tqdm,
                progress_callback=progress_callback,
            )
        elif isinstance(data_gen, BatchGeneratorBboxData):
            return self._predict_bboxes_data(
                n_bboxes_data_gen=data_gen,
                open_images_in_bboxes_data=open_images_in_data,
                disable_tqdm=disable_tqdm,
                progress_callback=progress_callback,
            )
        else:
            raise TypeError(f"Unknown type of data_gen: {type(data_gen)}")

    @property
    def class_names(self):
        return self.model.class_names
